import numpy as np
from astropy import units as u


class DarkData:
    """
    This class contains all data needed to perform the dark current subtraction
    step.
    """

    def __init__(self, dims=None, dark_model=None):
        """
        Creates a class to remove data model dependencies in the internals of
        the dark current code.  The data contained in this class comes from the
        dark reference data file.  If a dark data model is passed as an
        argument the DarkData uses it to create class arrays and set metadata.
        If a dark data model is None, then the DarkData arrays can still be
        created by using the 'dims' argument.  In that situation, no metadata
        will be contained in the class.  If both arguments are None, all arrays
        and metadata are set to None, requiring the instantiator of the class
        to set wanted values.

        Parameters
        ----------
        dims : tuple, optional
            A tuple of integers to describe the dimensions of the arrays used
            during the dark current step.  This argument is only used if the
            'dark_model' argument is None.  If a dark model is not available
            from which to create a DarkData class, but the dimensions of the
            data array are known, then 'dims' is used (the arrays data, groupdq,
            and err are assumed to have the same dimension).


        dark_model : data model, optional
            Input data model, assumed to be a JWST DarkModel like model.  If
            this argument is not None, the DarkData class will have arrasy and
            meta data set based on the arrays in the dark_model.
        """
        if dark_model is not None:
            if isinstance(dark_model.data, u.Quantity):
                self.data = dark_model.data.value
                self.err = dark_model.err.value
            else:
                self.data = dark_model.data
                self.err = dark_model.err
            self.groupdq = dark_model.dq

            self.exp_nframes = dark_model.meta.exposure.nframes
            self.exp_ngroups = dark_model.meta.exposure.ngroups
            self.exp_groupgap = dark_model.meta.exposure.groupgap

        elif dims is not None:
            self.data = np.zeros(dims, dtype=np.float32)
            self.groupdq = np.zeros(dims, dtype=np.uint32)
            self.err = np.zeros(dims, dtype=np.float32)

            self.exp_nframes = None
            self.exp_ngroups = None
            self.exp_groupgap = None

        else:
            self.data = None
            self.groupdq = None
            self.err = None

            self.exp_nframes = None
            self.exp_ngroups = None
            self.exp_groupgap = None

        self.save = False
        self.output_name = None


class ScienceData:
    def __init__(self, science_model=None):
        """
        A class containing all science data needed to subtract the dark current
        from the data.

        Parameters
        ----------
        science_model : data model, optional
            Input data model, assumed to be a JWST RampModel like model.  If
            this is None, then the class instantiator is responsible for
            populating the data.
        """
        if science_model is not None:
            if isinstance(science_model.data, u.Quantity):
                self.data = science_model.data.value
            else:
                self.data = science_model.data
            self.groupdq = science_model.groupdq
            self.pixeldq = science_model.pixeldq

            if isinstance(science_model.err, u.Quantity):
                self.err = science_model.err.value
            else:
                self.err = science_model.err

            self.exp_nframes = science_model.meta.exposure.nframes
            self.exp_groupgap = science_model.meta.exposure.groupgap
            try:  # JWST only
                self.exp_intstart = science_model.meta.exposure.integration_start
            except AttributeError:
                self.exp_intstart = None

            self.cal_step = None
        else:
            self.data = None
            self.groupdq = None
            self.pixeldq = None
            self.err = None

            self.exp_nframes = None
            self.exp_groupgap = None
            self.exp_intstart = None

            self.cal_step = None



# cython: language_level=3str


"""
This module contains all the functions needed to execute jump detection for the
    Castentano+22 ramp fitting algorithm

    The _ramp module contains the actual ramp fitting algorithm, this module
    contains a driver for the algorithm and detection of jumps/splitting ramps.

Structs
-------
Thresh : struct
    intercept - constant * log10(slope)
        - intercept : float
            The intercept of the jump threshold
        - constant : float
            The constant of the jump threshold

JumpFits : struct
    All the data on a given pixel's ramp fit with (or without) jump detection
        - average : RampFit
            The average of all the ramps fit for the pixel
        - jumps : vector[int]
            The indices of the resultants which were detected as jumps
        - fits : vector[RampFit]
            All of the fits for each ramp fit for the pixel
        - index : RampQueue
            The RampIndex representations corresponding to each fit in fits

Enums
-----
FixedOffsets : enum
    Enumerate the different pieces of information computed for jump detection
        which only depend on the read pattern.

PixelOffsets : enum
    Enumerate the different pieces of information computed for jump detection
        which only depend on the given pixel (independent of specific ramp).

JUMP_DET : value
    A the fixed value for the jump detection dq flag.

(Public) Functions
------------------
fill_fixed_values : function
    Pre-compute all the values needed for jump detection for a given read_pattern,
        this is independent of the pixel involved.

fit_jumps : function
    Compute all the ramps for a single pixel using the Casertano+22 algorithm
        with jump detection. This is a driver for the ramp fit algorithm in general
        meaning it automatically handles splitting ramps across dq flags in addition
        to splitting across detected jumps (if jump detection is turned on).
"""

from cython cimport boundscheck, cdivision, wraparound
from libc.math cimport NAN, fmaxf, isnan, log10, sqrtf
from libcpp cimport bool

from stcal.ramp_fitting.ols_cas22._jump cimport JUMP_DET, FixedOffsets, JumpFits, PixelOffsets, Thresh
from stcal.ramp_fitting.ols_cas22._ramp cimport RampFit, RampIndex, RampQueue, fit_ramp, init_ramps


@boundscheck(False)
@wraparound(False)
@cdivision(True)
cpdef inline float[:, :] fill_fixed_values(float[:, :] fixed,
                                           float[:] t_bar,
                                           float[:] tau,
                                           int[:] n_reads,
                                           int n_resultants):
    """
    Pre-compute all the values needed for jump detection which only depend on
        the read pattern.

    Parameters
    ----------
    fixed : float[:, :]
        A pre-allocated memoryview to store the pre-computed values in, its faster
        to allocate outside this function.
    t_bar : float[:]
        The average time for each resultant
    tau : float[:]
        The time variance for each resultant
    n_reads : int[:]
        The number of reads for each resultant
    n_resultants : int
        The number of resultants for the read pattern

    Returns
    -------
    [
        <t_bar[i+1] - t_bar[i]>,
        <t_bar[i+2] - t_bar[i]>,
        <t_bar[i+1] - t_bar[i]> ** 2,
        <t_bar[i+2] - t_bar[i]> ** 2,
        <(1/n_reads[i+1] + 1/n_reads[i])>,
        <(1/n_reads[i+2] + 1/n_reads[i])>,
        <(tau[i] + tau[i+1] - 2 * min(t_bar[i], t_bar[i+1]))>,
        <(tau[i] + tau[i+2] - 2 * min(t_bar[i], t_bar[i+2]))>,
    ]
    """
    # Cast the enum values into integers for indexing (otherwise compiler complains)
    #   These will be optimized out
    cdef int single_t_bar_diff = FixedOffsets.single_t_bar_diff
    cdef int double_t_bar_diff = FixedOffsets.double_t_bar_diff
    cdef int single_t_bar_diff_sqr = FixedOffsets.single_t_bar_diff_sqr
    cdef int double_t_bar_diff_sqr = FixedOffsets.double_t_bar_diff_sqr
    cdef int single_read_recip = FixedOffsets.single_read_recip
    cdef int double_read_recip = FixedOffsets.double_read_recip
    cdef int single_var_slope_val = FixedOffsets.single_var_slope_val
    cdef int double_var_slope_val = FixedOffsets.double_var_slope_val

    # Coerce division to be using floats
    cdef float num = 1

    cdef int i
    for i in range(n_resultants - 1):
        fixed[single_t_bar_diff, i] = t_bar[i + 1] - t_bar[i]
        fixed[single_t_bar_diff_sqr, i] = fixed[single_t_bar_diff, i] ** 2
        fixed[single_read_recip, i] = (num / n_reads[i + 1]) + (num / n_reads[i])
        fixed[single_var_slope_val, i] = tau[i + 1] + tau[i] - 2 * min(t_bar[i + 1], t_bar[i])

        if i < n_resultants - 2:
            fixed[double_t_bar_diff, i] = t_bar[i + 2] - t_bar[i]
            fixed[double_t_bar_diff_sqr, i] = fixed[double_t_bar_diff, i] ** 2
            fixed[double_read_recip, i] = (num / n_reads[i + 2]) + (num / n_reads[i])
            fixed[double_var_slope_val, i] = tau[i + 2] + tau[i] - 2 * min(t_bar[i + 2], t_bar[i])
        else:
            # Last double difference is undefined
            fixed[double_t_bar_diff, i] = NAN
            fixed[double_t_bar_diff_sqr, i] = NAN
            fixed[double_read_recip, i] = NAN
            fixed[double_var_slope_val, i] = NAN

    return fixed


@boundscheck(False)
@wraparound(False)
@cdivision(True)
cpdef inline float[:, :] _fill_pixel_values(float[:, :] pixel,
                                            float[:] resultants,
                                            float[:, :] fixed,
                                            float read_noise,
                                            int n_resultants):
    """
    Pre-compute all the values needed for jump detection which only depend on
        the a specific pixel (independent of the given ramp for a pixel).

    Parameters
    ----------
    pixel : float[:, :]
        A pre-allocated memoryview to store the pre-computed values in, its faster
        to allocate outside this function.
    resultants : float[:]
        The resultants for the pixel in question.
    fixed : float[:, :]
        The pre-computed fixed values for the read_pattern
    read_noise : float
        The read noise for the pixel
    n_resultants : int
        The number of resultants for the read_pattern

    Returns
    -------
    [
        <(resultants[i+1] - resultants[i])> / <(t_bar[i+1] - t_bar[i])>,
        <(resultants[i+2] - resultants[i])> / <(t_bar[i+2] - t_bar[i])>,
        read_noise**2 * <(1/n_reads[i+1] + 1/n_reads[i])>,
        read_noise**2 * <(1/n_reads[i+2] + 1/n_reads[i])>,
    ]
    """
    cdef int single_t_bar_diff = FixedOffsets.single_t_bar_diff
    cdef int double_t_bar_diff = FixedOffsets.double_t_bar_diff
    cdef int single_read_recip = FixedOffsets.single_read_recip
    cdef int double_read_recip = FixedOffsets.double_read_recip

    cdef int single_slope = PixelOffsets.single_local_slope
    cdef int double_slope = PixelOffsets.double_local_slope
    cdef int single_var = PixelOffsets.single_var_read_noise
    cdef int double_var = PixelOffsets.double_var_read_noise

    cdef float read_noise_sqr = read_noise ** 2

    cdef int i
    for i in range(n_resultants - 1):
        pixel[single_slope, i] = (resultants[i + 1] - resultants[i]) / fixed[single_t_bar_diff, i]
        pixel[single_var, i] = read_noise_sqr * fixed[single_read_recip, i]

        if i < n_resultants - 2:
            pixel[double_slope, i] = (resultants[i + 2] - resultants[i]) / fixed[double_t_bar_diff, i]
            pixel[double_var, i] = read_noise_sqr * fixed[double_read_recip, i]
        else:
            # The last double difference is undefined
            pixel[double_slope, i] = NAN
            pixel[double_var, i] = NAN

    return pixel


cdef inline float _threshold(Thresh thresh, float slope):
    """
    Compute jump threshold

    Parameters
    ----------
    thresh : Thresh
        threshold parameters struct
    slope : float
        slope of the ramp in question

    Returns
    -------
        intercept - constant * log10(slope)
    """
    slope = slope if slope > 1 else 1
    slope = slope if slope < 1e4 else 1e4

    return thresh.intercept - thresh.constant * log10(slope)


@boundscheck(False)
@wraparound(False)
@cdivision(True)
cdef inline float _correction(float[:] t_bar, RampIndex ramp, float slope):
    """
    Compute the correction factor for the variance used by a statistic

        - slope / (t_bar[end] - t_bar[start])

    Parameters
    ----------
    t_bar : float[:]
        The computed t_bar values for the ramp
    ramp : RampIndex
        Struct for start and end indices resultants for the ramp
    slope : float
        The computed slope for the ramp
    """

    cdef float diff = t_bar[ramp.end] - t_bar[ramp.start]

    return - slope / diff


@boundscheck(False)
@wraparound(False)
@cdivision(True)
cdef inline float _statstic(float local_slope,
                            float var_read_noise,
                            float t_bar_diff_sqr,
                            float var_slope_coeff,
                            float slope,
                            float correct):
    """
    Compute a single fit statistic
        delta / sqrt(var + correct)

    where:
        delta = local_slope - slope
        var = (var_read_noise + slope * var_slope_coeff) / t_bar_diff_sqr

        pre-computed:
            local_slope = (resultant[i + j]  - resultant[i]) / (t_bar[i + j] - t_bar[i])
            var_read_noise = read_noise ** 2 * (1/n_reads[i + j] + 1/n_reads[i])
            var_slope_coeff = tau[i + j] + tau[i] - 2 * min(t_bar[i + j], t_bar[i])
            t_bar_diff_sqr = (t_bar[i + j] - t_bar[i]) ** 2

    Parameters
    ----------
    local_slope : float
        The local slope the statistic is computed for
    float : var_read_noise
        The read noise variance for local_slope
    t_bar_diff_sqr : float
        The square difference for the t_bar corresponding to local_slope
    var_slope_coeff : float
        The slope variance coefficient for local_slope
    slope : float
        The computed slope for the ramp
    correct : float
        The correction factor needed

    Returns
    -------
        Create a single instance of the stastic for the given parameters
    """

    cdef float delta = local_slope - slope
    cdef float var = (var_read_noise + slope * var_slope_coeff) / t_bar_diff_sqr

    return delta / sqrtf(var + correct)


@boundscheck(False)
@wraparound(False)
cdef inline (int, float) _fit_statistic(float[:, :] pixel,
                                        float[:, :] fixed,
                                        float[:] t_bar,
                                        float slope,
                                        RampIndex ramp):
    """
    Compute the maximum index and its value over all fit statistics for a given
        ramp. Each index's stat is the max of the single and double difference
        statistics:
            all_stats = <max(single_stats), max(double_stats)>

    Parameters
    ----------
    pixel : float[:, :]
        The pre-computed fixed values for a given pixel
    fixed : float[:, :]
        The pre-computed fixed values for a given read_pattern
    t_bar : float[:, :]
        The average time for each resultant
    slope : float
        The computed slope for the ramp
    ramp : RampIndex
        Struct for start and end of ramp to fit

    Returns
    -------
        argmax(all_stats), max(all_stats)
    """
    # Cast the enum values into integers for indexing (otherwise compiler complains)
    #   These will be optimized out
    cdef int single_local_slope = PixelOffsets.single_local_slope
    cdef int double_local_slope = PixelOffsets.double_local_slope
    cdef int single_var_read_noise = PixelOffsets.single_var_read_noise
    cdef int double_var_read_noise = PixelOffsets.double_var_read_noise

    cdef int single_t_bar_diff_sqr = FixedOffsets.single_t_bar_diff_sqr
    cdef int double_t_bar_diff_sqr = FixedOffsets.double_t_bar_diff_sqr
    cdef int single_var_slope_val = FixedOffsets.single_var_slope_val
    cdef int double_var_slope_val = FixedOffsets.double_var_slope_val

    # Note that a ramp consisting of a single point is degenerate and has no
    #   fit statistic so we bail out here
    if ramp.start == ramp.end:
        return 0, NAN

    # Start computing fit statistics
    cdef float correct = _correction(t_bar, ramp, slope)

    # We are computing single and double differences of using the ramp's resultants.
    #    Each of these computations requires two points meaning that there are
    #    start - end - 1 possible differences. However, we cannot compute a double
    #    difference for the last point as there is no point after it. Therefore,
    #    We use this point's single difference as our initial guess for the fit
    #    statistic. Note that the fit statistic can technically be negative so
    #    this makes it much easier to compute a "lazy" max.
    cdef int index = ramp.end - 1
    cdef int argmax = ramp.end - ramp.start - 1
    cdef float max_stat = _statstic(pixel[single_local_slope, index],
                                    pixel[single_var_read_noise, index],
                                    fixed[single_t_bar_diff_sqr, index],
                                    fixed[single_var_slope_val, index],
                                    slope,
                                    correct)

    # Compute the rest of the fit statistics
    cdef float stat, stat1, stat2
    cdef int stat_index
    for stat_index, index in enumerate(range(ramp.start, ramp.end - 1)):
        # Compute max of single and double difference statistics
        stat1 = _statstic(pixel[single_local_slope, index],
                          pixel[single_var_read_noise, index],
                          fixed[single_t_bar_diff_sqr, index],
                          fixed[single_var_slope_val, index],
                          slope,
                          correct)
        stat2 = _statstic(pixel[double_local_slope, index],
                          pixel[double_var_read_noise, index],
                          fixed[double_t_bar_diff_sqr, index],
                          fixed[double_var_slope_val, index],
                          slope,
                          correct)
        stat = fmaxf(stat1, stat2)

        # If this is larger than the current max, update the max
        if stat > max_stat:
            max_stat = stat
            argmax = stat_index

    return argmax, max_stat


@boundscheck(False)
@wraparound(False)
@cdivision(True)
cdef inline JumpFits fit_jumps(float[:] resultants,
                               int[:] dq,
                               float read_noise,
                               float[:] t_bar,
                               float[:] tau,
                               int[:] n_reads,
                               int n_resultants,
                               float[:, :] fixed,
                               float[:, :] pixel,
                               Thresh thresh,
                               bool use_jump,
                               bool include_diagnostic):
    """
    Compute all the ramps for a single pixel using the Casertano+22 algorithm
        with jump detection.

    Parameters
    ----------
    resultants : float[:]
        The resultants for the pixel
    dq : int[:]
        The dq flags for the pixel. This is modified in place, so the external
        dq flag array will be modified as a side-effect.
    read_noise : float
        The read noise for the pixel.
    ramps : RampQueue
        RampQueue for initial ramps to fit for the pixel
        multiple ramps are possible due to dq flags
    t_bar : float[:]
        The average time for each resultant
    tau : float[:]
        The time variance for each resultant
    n_reads : int[:]
        The number of reads for each resultant
    n_resultants : int
        The number of resultants for the pixel
    fixed : float[:, :]
        The jump detection pre-computed values for a given read_pattern
    pixel : float[:, :]
        A pre-allocated array for the jump detection fixed values for the
        given pixel. This will be modified in place, it is passed in to avoid
        re-allocating it for each pixel.
    thresh : Thresh
        The threshold parameter struct for jump detection
    use_jump : bool
        Turn on or off jump detection.
    include_diagnostic : bool
        Turn on or off recording all the diaganostic information on the fit

    Returns
    -------
    RampFits struct of all the fits for a single pixel
    """
    # Find initial set of ramps
    cdef RampQueue ramps = init_ramps(dq, n_resultants)

    # Initialize algorithm
    cdef JumpFits ramp_fits
    cdef RampIndex ramp
    cdef RampFit ramp_fit

    ramp_fits.average.slope = 0
    ramp_fits.average.read_var = 0
    ramp_fits.average.poisson_var = 0

    cdef int argmax, jump0, jump1
    cdef float max_stat
    cdef float weight, total_weight = 0

    # Fill in the jump detection pre-compute values for a single pixel
    if use_jump:
        pixel = _fill_pixel_values(pixel, resultants, fixed, read_noise, n_resultants)

    # Run while the Queue is non-empty
    while not ramps.empty():
        # Remove top ramp of the stack to use
        ramp = ramps.back()
        ramps.pop_back()

        # Compute fit using the Casertano+22 algorithm
        ramp_fit = fit_ramp(resultants,
                            t_bar,
                            tau,
                            n_reads,
                            read_noise,
                            ramp)

        # Run jump detection if enabled
        if use_jump:
            argmax, max_stat = _fit_statistic(pixel,
                                              fixed,
                                              t_bar,
                                              ramp_fit.slope,
                                              ramp)

            # Note that when a "ramp" is a single point, _fit_statistic returns
            # a NaN for max_stat. Note that NaN > anything is always false so the
            # result drops through as desired.
            if max_stat > _threshold(thresh, ramp_fit.slope):
                # Compute jump point to create two new ramps
                #    This jump point corresponds to the index of the largest
                #    statistic:
                #        argmax = argmax(stats)
                #    These statistics are indexed relative to the
                #    ramp's range. Therefore, we need to add the start index
                #    of the ramp to the result.
                #
                jump0 = argmax + ramp.start

                # Note that because the resultants are averages of reads, but
                # jumps occur in individual reads, it is possible that the
                # jump is averaged down by the resultant with the actual jump
                # causing the computed jump to be off by one index.
                #     In the idealized case this is when the jump occurs near
                #     the start of the resultant with the jump. In this case,
                #     the statistic for the resultant will be maximized at
                #     index - 1 rather than index. This means that we have to
                #     remove argmax(stats) + 1 as it is also a possible jump.
                #     This case is difficult to distinguish from the case where
                #     argmax(stats) does correspond to the jump resultant.
                #     Therefore, we just remove both possible resultants from
                #     consideration.
                jump1 = jump0 + 1

                # Update the dq flags
                dq[jump0] = JUMP_DET
                dq[jump1] = JUMP_DET

                # Record jump diagnostics
                if include_diagnostic:
                    ramp_fits.jumps.push_back(jump0)
                    ramp_fits.jumps.push_back(jump1)

                # The two resultant indices need to be skipped, therefore
                # the two
                # possible new ramps are:
                #     RampIndex(ramp.start, jump0 - 1)
                #     RampIndex(jump1 + 1, ramp.end)
                # This is because the RampIndex contains the index of the
                # first and last resulants in the sub-ramp it describes.
                #    Note: The algorithm works via working over the sub-ramps
                #    backward in time. Therefore, since we are using a stack,
                #    we need to add the ramps in the time order they were
                #    observed in. This results in the last observation ramp
                #    being the top of the stack; meaning that,
                #    it will be the next ramp handled.

                if jump0 > ramp.start:
                    # Note that when jump0 == ramp.start, we have detected a
                    # jump in the first resultant of the ramp. This means
                    # there is no sub-ramp before jump0.
                    #    Also, note that this will produce bad results as
                    #    the ramp indexing will go out of bounds. So it is
                    #    important that we exclude it.
                    # Note that jump0 < ramp.start is not possible because
                    # the argmax is always >= 0
                    ramps.push_back(RampIndex(ramp.start, jump0 - 1))

                if jump1 < ramp.end:
                    # Note that if jump1 == ramp.end, we have detected a
                    # jump in the last resultant of the ramp. This means
                    # there is no sub-ramp after jump1.
                    #    Also, note that this will produce bad results as
                    #    the ramp indexing will go out of bounds. So it is
                    #    important that we exclude it.
                    # Note that jump1 > ramp.end is technically possible
                    # however in those potential cases it will draw on
                    # resultants which are not considered part of the ramp
                    # under consideration. Therefore, we have to exclude all
                    # of those values.
                    ramps.push_back(RampIndex(jump1 + 1, ramp.end))

                # Skip recording the ramp as it has a detected jump
                continue

        # Start recording the the fit (no jump detected)

        # Record the diagnositcs
        if include_diagnostic:
            ramp_fits.fits.push_back(ramp_fit)
            ramp_fits.index.push_back(ramp)

        # Start computing the averages using a lazy process
        #    Note we do not do anything in the NaN case for degenerate ramps
        if not isnan(ramp_fit.slope):
            # protect weight against the extremely unlikely case of a zero
            # variance
            weight = 0 if ramp_fit.read_var == 0 else 1 / ramp_fit.read_var
            total_weight += weight

            ramp_fits.average.slope += weight * ramp_fit.slope
            ramp_fits.average.read_var += weight**2 * ramp_fit.read_var
            ramp_fits.average.poisson_var += weight**2 * ramp_fit.poisson_var

    # Finish computing averages using the lazy process
    ramp_fits.average.slope /= total_weight if total_weight != 0 else 1
    ramp_fits.average.read_var /= total_weight**2 if total_weight != 0 else 1
    ramp_fits.average.poisson_var /= total_weight**2 if total_weight != 0 else 1

    # Multiply poisson term by flux, (no negative fluxes)
    ramp_fits.average.poisson_var *= max(ramp_fits.average.slope, 0)

    return ramp_fits



import numpy as np
import pytest

from stcal.ramp_fitting.ramp_fit import ramp_fit_class, ramp_fit_data

test_dq_flags = {
    "GOOD": 0,  # Good pixel.
    "DO_NOT_USE": 2**0,  # Bad pixel. Do not use.
    "SATURATED": 2**1,  # Pixel saturated during exposure.
    "JUMP_DET": 2**2,  # Jump detected during exposure.
    "CHARGELOSS": 2**7,   # Charge migration (was RESERVED_4)
    "NO_GAIN_VALUE": 2**19,  # Gain cannot be measured.
    "UNRELIABLE_SLOPE": 2**24,  # Slope variance large (i.e., noisy pixel).
}

GOOD = test_dq_flags["GOOD"]
DO_NOT_USE = test_dq_flags["DO_NOT_USE"]
JUMP_DET = test_dq_flags["JUMP_DET"]
SATURATED = test_dq_flags["SATURATED"]
NO_GAIN_VALUE = test_dq_flags["NO_GAIN_VALUE"]
UNRELIABLE_SLOPE = test_dq_flags["UNRELIABLE_SLOPE"]
CHRGL = test_dq_flags["CHARGELOSS"]

DELIM = "-" * 70


def setup_inputs(dims, gain, rnoise, group_time, frame_time):
    """
    Creates test data for testing.  All ramp data is zero.

    Parameters
    ----------
    dims: tuple
        Four dimensions (nints, ngroups, nrows, ncols)

    gain: float
        Gain noise

    rnoise: float
        Read noise

    group_time: float
        Group time

    frame_time: float
        Frame time

    Return
    ------
    ramp_class: RampClass
        A RampClass with all zero data.

    gain: ndarray
        A 2-D array for gain noise for each pixel.

    rnoise: ndarray
        A 2-D array for read noise for each pixel.
    """
    nints, ngroups, nrows, ncols = dims

    ramp_class = ramp_fit_class.RampData()  # Create class

    # Create zero arrays according to dimensions
    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    err = np.ones(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    groupdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint8)
    pixeldq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    dark_current = np.zeros(shape=(nrows, ncols), dtype=np.float32)


    # Set clas arrays
    ramp_class.set_arrays(data, err, groupdq, pixeldq, average_dark_current=dark_current)

    # Set class meta
    ramp_class.set_meta(
        name="MIRI",
        frame_time=frame_time,
        group_time=group_time,
        groupgap=0,
        nframes=1,
        drop_frames1=0,
    )

    # Set class data quality flags
    ramp_class.set_dqflags(test_dq_flags)

    # Set noise arrays
    gain = np.ones(shape=(nrows, ncols), dtype=np.float64) * gain
    rnoise = np.full((nrows, ncols), rnoise, dtype=np.float32)

    return ramp_class, gain, rnoise


# -----------------------------------------------------------------------------


def test_one_group_small_buffer():
    """
    Checks to make sure if a single group is used, it works.
    TODO: It does not work.  GLS needs to be modified to work edge cases.
    """
    nints, ngroups, nrows, ncols = 1, 1, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_data.data[0, 0, 50, 50] = 10.0

    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    data = slopes[0]
    tol = 1.0e-6
    np.testing.assert_allclose(data[50, 50], 10.0, tol)


def test_two_integrations():
    """
    A test to see if GLS is correctly combining integrations.
    """
    nints, ngroups, nrows, ncols = 2, 11, 1, 1
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 1, 5
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    row, col = 0, 0

    ramp = np.asarray([x * 100 for x in range(11)])
    ramp_data.data[0, :, row, col] = ramp
    ramp_data.data[1, :, row, col] = ramp * 2

    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    np.testing.assert_allclose(slopes[0][row, col], 133.3377685, 1e-6)


def test_one_group_two_integrations():
    """
    Test for multiple integrations with only one group.
    """
    nints, ngroups, nrows, ncols = 2, 1, 1, 1
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_data.data[0, 0, 0, 0] = 10.0
    ramp_data.data[1, 0, 0, 0] = 11.0

    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    data = slopes[0]

    check = 10.5
    np.testing.assert_allclose(data[0, 0], check, 1e-6)


def test_nocrs_noflux():
    """
    Make sure no data returns all zeros.
    """
    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    assert np.max(slopes[0]) == 0
    assert np.min(slopes[0]) == 0


@pytest.mark.skip(reason="Getting all NaN's, but expecting all zeros.")
def test_nocrs_noflux_firstrows_are_nan():
    """
    The 12 rows are set to NaN.  Not sure why this is expected to return
    all zeros.
    """
    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0:, 0:12, :] = np.nan

    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    assert np.max(slopes[0]) == 0
    assert np.min(slopes[0]) == 0


def test_error_when_frame_time_not_set():
    """
    Frame time is needed, so make sure an exception gets raised.
    """
    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0:, 0:12, :] = np.nan

    ramp_data.frame_time = None  # Must be set

    save_opt, algo, ncores = False, "GLS", "none"
    with pytest.raises(UnboundLocalError):
        slopes, cube, ols_opt, gls_opt = ramp_fit_data(
            ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
        )


def test_five_groups_two_integrations_Poisson_noise_only():
    """
    Multi-group ramp, with multi-integrations, with large poisson noise.
    """
    nints, ngroups, nrows, ncols = 2, 5, 1, 1
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 7, 2000
    group_time, frame_time = 3.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    row, col = 0, 0
    ramp_data.data[0, 0, row, col] = 10.0
    ramp_data.data[0, 1, row, col] = 15.0
    ramp_data.data[0, 2, row, col] = 25.0
    ramp_data.data[0, 3, row, col] = 33.0
    ramp_data.data[0, 4, row, col] = 60.0
    ramp_data.data[1, 0, row, col] = 10.0
    ramp_data.data[1, 1, row, col] = 15.0
    ramp_data.data[1, 2, row, col] = 25.0
    ramp_data.data[1, 3, row, col] = 33.0
    ramp_data.data[1, 4, row, col] = 160.0

    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    out_slope = slopes[0][row, col]
    deltaDN1 = 50
    deltaDN2 = 150
    check = (deltaDN1 + deltaDN2) / 2.0

    np.testing.assert_allclose(out_slope, check, 75.0, 1e-6)


def test_bad_gain_values():
    """
    Test for bad gain values where gain values are negative
    and NaN.
    """
    nints, ngroups, nrows, ncols = 1, 5, 10, 11
    r1, c1, r2, c2 = 3, 3, 7, 7
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 7, 2000
    group_time, frame_time = 3.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    gain2d[r1, c1] = -10
    gain2d[r2, c2] = np.nan

    # save_opt, algo, ncores = False, "OLS", "none"
    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    # data, dq, var_poisson, var_rnoise, err = slopes
    data, dq, err = slopes
    flag_check = NO_GAIN_VALUE | DO_NOT_USE

    assert dq[r1, c1] == flag_check
    assert dq[r2, c2] == flag_check

    # These asserts are wrong for some reason
    assert np.max(data) == 0
    assert np.min(data) == 0


def test_simple_ramp():
    """
    Here given a 10 group ramp with an exact slope of 20/group.
    The output slope should be 20.
    """
    nints, ngroups, nrows, ncols = 1, 10, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 7, 2000
    group_time, frame_time = 3.0, 3

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp = np.array(list(range(ngroups))) * 20 + 10
    ramp_data.data[0, :, 50, 50] = ramp

    save_opt, algo, ncores = False, "GLS", "none"
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, 512, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    check = 20.0 / 3
    tol = 1.0e-5
    np.testing.assert_allclose(answer, check, tol)


def test_read_noise_only_fit():
    """
    Checks ramp fit GLS against polyfit, but it is slightly off.
    """
    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 50, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_arr = [10.0, 15.0, 25.0, 33.0, 60.0]
    ramp_data.data[0, :, 50, 50] = np.array(ramp_arr)

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    xvalues = np.arange(5) * 1.0
    yvalues = np.array(ramp_arr)
    coeff = np.polyfit(xvalues, yvalues, 1)
    answer = slopes[0][50, 50]
    check = coeff[0]
    tol = 1.0e-2
    # print(f"answer = {answer}")         # 11.78866004
    # print(f"check = {check}")     # 11.79999999
    np.testing.assert_allclose(answer, check, tol)


@pytest.mark.skip(reason="GLS not sure what expected value is.")
def test_photon_noise_only_fit():
    """ """
    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 1, 1000
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_arr = [10.0, 15.0, 25.0, 33.0, 60.0]
    ramp_data.data[0, :, 50, 50] = np.array(ramp_arr)

    check = (ramp_data.data[0, 4, 50, 50] - ramp_data.data[0, 0, 50, 50]) / 4.0

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    tol = 1.0e-2
    # print(f"answer = {answer}")         #  8.6579208
    # print(f"check = {check}")     # 12.5
    np.testing.assert_allclose(answer, check, tol)


@pytest.mark.skip(reason="GLS not sure what expected value is.")
def test_photon_noise_only_bad_last_group():
    """ """
    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 1, 1000
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 15.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 33.0
    ramp_data.data[0, 4, 50, 50] = 60.0

    check = (ramp_data.data[0, 3, 50, 50] - ramp_data.data[0, 0, 50, 50]) / 3.0

    ramp_data.groupdq[0, 4, :, :] = DO_NOT_USE

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    tol = 1.0e-2
    # print(f"answer = {answer}")         # 8.6579208
    # print(f"check = {check}")     # 7.6666666
    np.testing.assert_allclose(answer, check, tol)


@pytest.mark.skip(reason="GLS not sure what expected value is.")
def test_photon_noise_with_unweighted_fit():
    """ """

    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 1, 1000
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 15.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 33.0
    ramp_data.data[0, 4, 50, 50] = 60.0

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "unweighted", ncores, test_dq_flags
    )

    xvalues = np.arange(5) * 1.0
    yvalues = np.array([10, 15, 25, 33, 60])
    coeff = np.polyfit(xvalues, yvalues, 1)
    check = coeff[0]
    answer = slopes[0][50, 50]
    tol = 1.0e-5
    # print(f"answer = {answer}")         #  8.6579208
    # print(f"check = {check}")     # 11.7999999
    np.testing.assert_allclose(answer, check, tol)


def test_two_groups_fit():
    """
    Ensure pixels with two group ramps and saturated groups get their
    final DQ flags set properly.
    """
    nints, ngroups, nrows, ncols = 1, 2, 1, 3
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0, 0, 0, 0] = 10.0
    ramp_data.data[0, 1, 0, 0] = 15.0
    ramp_data.data[0, 0, 0, 1] = 20.0
    ramp_data.data[0, 0, 0, 2] = 200.0
    ramp_data.data[0, 1, 0, 2] = 600.0
    check = ramp_data.data[0, 1, 0, 0] - ramp_data.data[0, 0, 0, 0]

    ramp_data.drop_frames1 = 0
    # 2nd group is saturated
    ramp_data.groupdq[0, 1, 0, 1] = SATURATED

    # 1st group is saturated
    ramp_data.groupdq[0, 0, 0, 2] = SATURATED
    ramp_data.groupdq[0, 1, 0, 2] = SATURATED

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    ans_data = slopes[0][0, 0]
    ans_dq = slopes[1]
    tol = 1.0e-5
    np.testing.assert_allclose(ans_data, check, tol)

    assert ans_dq[0, 0] == GOOD
    assert ans_dq[0, 1] == UNRELIABLE_SLOPE
    assert ans_dq[0, 2] == SATURATED | DO_NOT_USE


def test_four_groups_oneCR_orphangroupatend_fit():
    """ """
    nints, ngroups, nrows, ncols = 1, 4, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 15.0
    ramp_data.data[0, 2, 50, 50] = 20.0
    ramp_data.data[0, 3, 50, 50] = 145.0

    ramp_data.groupdq[0, 3, 50, 50] = JUMP_DET

    check = ramp_data.data[0, 1, 50, 50] - ramp_data.data[0, 0, 50, 50]

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    tol = 1.0e-6
    np.testing.assert_allclose(answer, check, tol)


def test_four_groups_two_CRs_at_end():
    """ """

    nints, ngroups, nrows, ncols = 1, 4, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 15.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 145.0
    check = ramp_data.data[0, 1, 50, 50] - ramp_data.data[0, 0, 50, 50]

    ramp_data.groupdq[0, 2, 50, 50] = JUMP_DET
    ramp_data.groupdq[0, 3, 50, 50] = JUMP_DET

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    tol = 1.0e-6
    np.testing.assert_allclose(answer, check, tol)


@pytest.mark.skip(reason="GLS code does not [yet] handle all groups as jump.")
def test_four_groups_four_CRs():
    """ """
    nints, ngroups, nrows, ncols = 1, 10, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 15.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 145.0

    ramp_data.groupdq[0, 0, 50, 50] = JUMP_DET
    ramp_data.groupdq[0, 1, 50, 50] = JUMP_DET
    ramp_data.groupdq[0, 2, 50, 50] = JUMP_DET
    ramp_data.groupdq[0, 3, 50, 50] = JUMP_DET

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    check = 0
    tol = 1.0e-6
    # print(f"answer = {answer}")
    # print(f"check = {check}")
    np.testing.assert_allclose(answer, check, tol)


@pytest.mark.skip(reason="GLS code does not [yet] handle only one good group.")
def test_four_groups_three_CRs_at_end():
    """ """
    nints, ngroups, nrows, ncols = 1, 4, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 15.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 145.0

    ramp_data.groupdq[0, 1, 50, 50] = JUMP_DET
    ramp_data.groupdq[0, 2, 50, 50] = JUMP_DET
    ramp_data.groupdq[0, 3, 50, 50] = JUMP_DET

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    check = 10.0
    tol = 1.0e-6
    # print(f"answer = {answer}")
    # print(f"check = {check}")
    np.testing.assert_allclose(answer, check, tol)


def test_four_groups_CR_causes_orphan_1st_group():
    """ """
    nints, ngroups, nrows, ncols = 1, 4, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10000, 0.01
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 125.0
    ramp_data.data[0, 2, 50, 50] = 145.0
    ramp_data.data[0, 3, 50, 50] = 165.0

    ramp_data.groupdq[0, 1, 50, 50] = JUMP_DET

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    check = 20.0
    tol = 1.0e-6
    np.testing.assert_allclose(answer, check, tol)


def test_one_group_fit():
    """ """
    nints, ngroups, nrows, ncols = 1, 1, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 1
    group_time, frame_time = 1.0, 1

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)
    ramp_data.data[0, 0, 50, 50] = 10.0

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    check = 10.0
    tol = 1.0e-6
    np.testing.assert_allclose(answer, check, tol)


@pytest.mark.skip(reason="GLS not sure what expected value is.")
def test_two_groups_unc():
    """ """
    deltaDN = 5  # TODO: Not sure wha this is supposed to be.
    nints, ngroups, nrows, ncols = 1, 2, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 10, 2
    group_time, frame_time = 3.0, 3.0

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 10.0 + deltaDN

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[2][50, 50]
    check = np.sqrt((deltaDN / gain) / group_time**2 + (rnoise**2 / group_time**2))
    tol = 1.0e-6
    # print(f"answer = {answer}")
    # print(f"check = {check}")
    np.testing.assert_allclose(answer, check, tol)


@pytest.mark.skip(reason="GLS does not comopute VAR_XXX arrays.")
def test_five_groups_unc():
    """ """
    """
        grouptime=3.0
        # deltaDN = 5
        ingain = 2
        inreadnoise =7
        ngroups=5
    ramp_data, gdq, rnModel, pixdq, err, gain = setup_inputs(ngroups=ngroups,
            gain=ingain, readnoise=inreadnoise, deltatime=grouptime)
    """
    nints, ngroups, nrows, ncols = 1, 5, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 7, 2
    group_time, frame_time = 3.0, 3

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    ramp_data.data[0, 0, 50, 50] = 10.0
    ramp_data.data[0, 1, 50, 50] = 15.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 33.0
    ramp_data.data[0, 4, 50, 50] = 60.0

    # check = np.median(np.diff(ramp_data.data[0,:,50,50])) / grouptime

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    """
    # Not sure what to do with this test.  The VAR_XXX arrays don't get
    # computed in GLS.

        # out_slope=slopes[0].data[500, 500]
        # deltaDN = 50
        delta_time = (ngroups - 1) * grouptime
        # delta_electrons = median_slope * ingain *delta_time
        single_sample_readnoise = np.float64(inreadnoise / np.sqrt(2))
        np.testing.assert_allclose(slopes[0].var_poisson[50,50],
            ((median_slope)/(ingain*delta_time)), 1e-6)
        np.testing.assert_allclose(slopes[0].var_rnoise[50,50],
            (12 * single_sample_readnoise**2/(ngroups * (ngroups**2 - 1) * grouptime**2)),  1e-6)
        np.testing.assert_allclose(slopes[0].err[50,50],
            np.sqrt(slopes[0].var_poisson[50,50]  + slopes[0].var_rnoise[50,50] ),  1e-6)
    """


@pytest.mark.skip(reason="GLS doesn't produce the optional results product, yet.")
def test_oneCR_10_groups_combination():
    """ """
    nints, ngroups, nrows, ncols = 1, 10, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    rnoise, gain = 7, 200
    group_time, frame_time = 3.0, 3

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    # two segments perfect fit, second segment has twice the slope
    ramp_data.data[0, 0, 50, 50] = 15.0
    ramp_data.data[0, 1, 50, 50] = 20.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 30.0
    ramp_data.data[0, 4, 50, 50] = 35.0
    ramp_data.data[0, 5, 50, 50] = 140.0
    ramp_data.data[0, 6, 50, 50] = 150.0
    ramp_data.data[0, 7, 50, 50] = 160.0
    ramp_data.data[0, 8, 50, 50] = 170.0
    ramp_data.data[0, 9, 50, 50] = 180.0

    ramp_data.groupdq[0, 5, 50, 50] = JUMP_DET

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    answer = slopes[0][50, 50]
    print(f"answer = {answer}")

    # TODO Need to add the optional results product to GLS

    """
    segment_groups  = 5
    single_sample_readnoise = np.float64(inreadnoise / np.sqrt(2))

    #check that the segment variance is as expected
    np.testing.assert_allclose(opt_model.var_rnoise[0,0,50,50],
        (12.0 * single_sample_readnoise**2 \
         / (segment_groups * (segment_groups**2 - 1) * grouptime**2)),
        rtol=1e-6)
    # check the combined slope is the average of the two segments since
    # they have the same number of groups
    np.testing.assert_allclose(slopes.data[50, 50], 2.5,rtol=1e-5)

    #check that the slopes of the two segments are correct
    np.testing.assert_allclose(opt_model.slope[0,0,50, 50], 5/3.0,rtol=1e-5)
    np.testing.assert_allclose(opt_model.slope[0,1,50, 50], 10/3.0,rtol=1e-5)
    """


@pytest.mark.skip(reason="GLS doesn't produce the optional results product, yet.")
def test_oneCR_10_groups_combination_noisy2ndSegment():
    """ """
    nints, ngroups, nrows, ncols = 1, 10, 103, 102
    dims = (nints, ngroups, nrows, ncols)
    # use large gain to show that Poisson noise doesn't affect the recombination
    rnoise, gain = 7, 200
    group_time, frame_time = 3.0, 3

    ramp_data, gain2d, rnoise2d = setup_inputs(dims, gain, rnoise, group_time, frame_time)

    # two segments perfect fit, second segment has twice the slope
    ramp_data.data[0, 0, 50, 50] = 15.0
    ramp_data.data[0, 1, 50, 50] = 20.0
    ramp_data.data[0, 2, 50, 50] = 25.0
    ramp_data.data[0, 3, 50, 50] = 30.0
    ramp_data.data[0, 4, 50, 50] = 35.0
    ramp_data.data[0, 5, 50, 50] = 135.0
    ramp_data.data[0, 6, 50, 50] = 155.0
    ramp_data.data[0, 7, 50, 50] = 160.0
    ramp_data.data[0, 8, 50, 50] = 168.0
    ramp_data.data[0, 9, 50, 50] = 180.0

    ramp_data.groupdq[0, 5, 50, 50] = JUMP_DET

    save_opt, algo, ncores, bufsize = False, "GLS", "none", 1024 * 30000
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise2d, gain2d, algo, "optimal", ncores, test_dq_flags
    )

    """
    avg_slope = (opt_model.slope[0,0,50,50] + opt_model.slope[0,1,50,50])/2.0
    # even with noiser second segment, final slope should be just the average
    # since they have the same number of groups
    np.testing.assert_allclose(slopes.data[50, 50], avg_slope,rtol=1e-5)
    """



# cython: language_level=3str

"""
This module contains all the functions needed to execute the Casertano+22 ramp
    fitting algorithm on its own without jump detection.

    The _jump module contains a driver function which calls the `fit_ramp` function
    from this module iteratively. This evvetively handles dq flags and detected
    jumps simultaneously.

Structs
-------
RampIndex : struct
    - start : int
        Index of the first resultant in the ramp
    - end : int
        Index of the last resultant in the ramp (so indexing of ramp requires end + 1)

RampFit : struct
    - slope : float
        The slope fit to the ramp
    - read_var : float
        The read noise variance for the fit
    - poisson_var : float
        The poisson variance for the fit

RampQueue : vector[RampIndex]
    Vector of RampIndex objects (convenience typedef)

Classes
-------
ReadPattern :
    Container class for all the metadata derived from the read pattern, this
    is just a temporary object to allow us to return multiple memory views from
    a single function.

(Public) Functions
------------------
init_ramps : function
    Create the initial ramp "queue" for each pixel in order to handle any initial
    "dq" flags passed in from outside.

from_read_pattern : function
    Derive the input data from the the read pattern
        This is faster than using __init__ or __cinit__ to construct the object with
        these calls.

fit_ramps : function
    Implementation of running the Casertano+22 algorithm on a (sub)set of resultants
    listed for a single pixel
"""
import numpy as np

cimport numpy as cnp
from cython cimport boundscheck, cdivision, cpow, wraparound
from libc.math cimport INFINITY, NAN, fabs, fmaxf, sqrt
from libcpp.vector cimport vector

from stcal.ramp_fitting.ols_cas22._ramp cimport RampFit, RampIndex, RampQueue, ReadPattern

# Initialize numpy for cython use in this module
cnp.import_array()


cdef class ReadPattern:
    """
    Class to contain the read pattern derived metadata
        This exists only to allow us to output multiple memory views at the same time
        from the same cython function. This is needed because neither structs nor unions
        can contain memory views.

        In the case of this code memory views are the fastest "safe" array data structure.
        This class will immediately be unpacked into raw memory views, so that we avoid
        any further overhead of switching between python and cython.

    Attributes:
    ----------
    t_bar : np.ndarray[float_t, ndim=1]
        The mean time of each resultant
    tau : np.ndarray[float_t, ndim=1]
        The variance in time of each resultant
    n_reads : np.ndarray[cnp.int32_t, ndim=1]
        The number of reads in each resultant
    """

    def _to_dict(ReadPattern self):
        """
        This is a private method to convert the ReadPattern object to a dictionary,
            so that attributes can be directly accessed in python. Note that this
            is needed because class attributes cannot be accessed on cython classes
            directly in python. Instead they need to be accessed or set using a
            python compatible method. This method is a pure puthon method bound
            to to the cython class and should not be used by any cython code, and
            only exists for testing purposes.
        """
        return dict(t_bar=np.array(self.t_bar, dtype=np.float32),
                    tau=np.array(self.tau, dtype=np.float32),
                    n_reads=np.array(self.n_reads, dtype=np.int32))


@boundscheck(False)
@wraparound(False)
@cdivision(True)
cpdef ReadPattern from_read_pattern(list[list[int]] read_pattern, float read_time, int n_resultants):
    """
    Derive the input data from the the read pattern
        This is faster than using __init__ or __cinit__ to construct the object with
        these calls.

    Parameters
    ----------
    read pattern: list[list[int]]
        read pattern for the image
    read_time : float
        Time to perform a readout.
    n_resultants : int
        Number of resultants in the image

    Returns
    -------
    ReadPattern
        Contains:
        - t_bar
        - tau
        - n_reads
    """

    cdef ReadPattern data = ReadPattern()
    data.t_bar = np.empty(n_resultants, dtype=np.float32)
    data.tau = np.empty(n_resultants, dtype=np.float32)
    data.n_reads = np.empty(n_resultants, dtype=np.int32)

    cdef int index, n_reads
    cdef list[int] resultant
    for index, resultant in enumerate(read_pattern):
        n_reads = len(resultant)

        data.n_reads[index] = n_reads
        data.t_bar[index] = read_time * np.mean(resultant)
        data.tau[index] = (np.sum((2 * (n_reads - np.arange(n_reads)) - 1) * resultant) *
                           read_time / n_reads**2)

    return data


@boundscheck(False)
@wraparound(False)
cpdef inline RampQueue init_ramps(int[:] dq, int n_resultants):
    """
    Create the initial ramp "queue" for each pixel
        if dq[index_resultant, index_pixel] == 0, then the resultant is in a ramp
        otherwise, the resultant is not in a ramp

    Parameters
    ----------
    dq : int[n_resultants]
        DQ array
    n_resultants : int
        Number of resultants

    Returns
    -------
    RampQueue
        vector of RampIndex objects
            - vector with entry for each ramp found (last entry is last ramp found)
            - RampIndex with start and end indices of the ramp in the resultants
    """
    cdef RampQueue ramps = RampQueue()

    # Note: if start/end are -1, then no value has been assigned
    # ramp.start == -1 means we have not started a ramp
    # dq[index_resultant, index_pixel] == 0 means resultant is in ramp
    cdef RampIndex ramp = RampIndex(-1, -1)
    cdef int index_resultant
    for index_resultant in range(n_resultants):
        if ramp.start == -1:
            # Looking for the start of a ramp
            if dq[index_resultant] == 0:
                # We have found the start of a ramp!
                ramp.start = index_resultant
            else:
                # This is not the start of the ramp yet
                continue
        else:
            # Looking for the end of a ramp
            if dq[index_resultant] == 0:
                # This pixel is in the ramp do nothing
                continue
            else:
                # This pixel is not in the ramp
                # => index_resultant - 1 is the end of the ramp
                ramp.end = index_resultant - 1

                # Add completed ramp to stack and reset ramp
                ramps.push_back(ramp)
                ramp = RampIndex(-1, -1)

    # Handle case where last resultant is in ramp (so no end has been set)
    if ramp.start != -1 and ramp.end == -1:
        # Last resultant is end of the ramp => set then add to stack
        ramp.end = n_resultants - 1
        ramps.push_back(ramp)

    return ramps

# Keeps the static type checker/highlighter happy this has no actual effect
ctypedef float[6] _row

# Casertano+2022, Table 2
cdef _row[2] _PTABLE = [[-INFINITY, 5,   10, 20, 50, 100],
                        [0,         0.4, 1,  3,  6,  10]]


@boundscheck(False)
@wraparound(False)
cdef inline float _get_power(float signal):
    """
    Return the power from Casertano+22, Table 2

    Parameters
    ----------
    signal: float
        signal from the resultants

    Returns
    -------
    signal power from Table 2
    """
    cdef int i
    for i in range(6):
        if signal < _PTABLE[0][i]:
            return _PTABLE[1][i - 1]

    return _PTABLE[1][i]


@boundscheck(False)
@wraparound(False)
@cdivision(True)
cdef inline RampFit fit_ramp(float[:] resultants_,
                             float[:] t_bar_,
                             float[:] tau_,
                             int[:] n_reads_,
                             float read_noise,
                             RampIndex ramp):
    """
    Fit a single ramp using Casertano+22 algorithm.

    Parameters
    ----------
    resultants_ : float[:]
        All of the resultants for the pixel
    t_bar_ : float[:]
        All the t_bar values
    tau_ : float[:]
        All the tau values
    n_reads_ : int[:]
        All the n_reads values
    read_noise : float
        The read noise for the pixel
    ramp : RampIndex
        Struct for start and end of ramp to fit

    Returns
    -------
    RampFit
        struct containing
        - slope
        - read_var
        - poisson_var
    """
    cdef int n_resultants = ramp.end - ramp.start + 1

    # Special case where there is no or one resultant, there is no fit and
    # we bail out before any computations.
    #    Note that in this case, we cannot compute the slope or the variances
    #    because these computations require at least two resultants. Therefore,
    #    this case is degernate and we return NaNs for the values.
    if n_resultants <= 1:
        return RampFit(NAN, NAN, NAN)

    # Compute the fit
    cdef int i = 0, j = 0

    # Setup data for fitting (work over subset of data) to make things cleaner
    #    Recall that the RampIndex contains the index of the first and last
    #    index of the ramp. Therefore, the Python slice needed to get all the
    #    data within the ramp is:
    #         ramp.start:ramp.end + 1
    cdef float[:] resultants = resultants_[ramp.start:ramp.end + 1]
    cdef float[:] t_bar = t_bar_[ramp.start:ramp.end + 1]
    cdef float[:] tau = tau_[ramp.start:ramp.end + 1]
    cdef int[:] n_reads = n_reads_[ramp.start:ramp.end + 1]

    # Compute mid point time
    cdef int end = n_resultants - 1
    cdef float t_bar_mid = (t_bar[0] + t_bar[end]) / 2

    # Casertano+2022 Eq. 44
    #    Note we've departed from Casertano+22 slightly;
    #    there s is just resultants[ramp.end].  But that doesn't seem good if, e.g.,
    #    a CR in the first resultant has boosted the whole ramp high but there
    #    is no actual signal.
    cdef float power = fmaxf(resultants[end] - resultants[0], 0)
    power = power / sqrt(read_noise**2 + power)
    power = _get_power(power)

    # It's easy to use up a lot of dynamic range on something like
    # (tbar - tbarmid) ** 10.  Rescale these.
    cdef float t_scale = (t_bar[end] - t_bar[0]) / 2
    t_scale = 1 if t_scale == 0 else t_scale

    # Initialize the fit loop
    #   it is faster to generate a c++ vector than a numpy array
    cdef vector[float] weights = vector[float](n_resultants)
    cdef vector[float] coeffs = vector[float](n_resultants)
    cdef RampFit ramp_fit = RampFit(0, 0, 0)
    cdef float f0 = 0, f1 = 0, f2 = 0
    cdef float coeff

    # Issue when tbar[] == tbarmid causes exception otherwise
    with cpow(True):
        for i in range(n_resultants):
            # Casertano+22, Eq. 45
            weights[i] = ((((1 + power) * n_reads[i]) / (1 + power * n_reads[i])) *
                          fabs((t_bar[i] - t_bar_mid) / t_scale) ** power)

            # Casertano+22 Eq. 35
            f0 += weights[i]
            f1 += weights[i] * t_bar[i]
            f2 += weights[i] * t_bar[i]**2

    # Casertano+22 Eq. 36
    cdef float det = f2 * f0 - f1 ** 2
    if det == 0:
        return ramp_fit

    for i in range(n_resultants):
        # Casertano+22 Eq. 37
        coeff = (f0 * t_bar[i] - f1) * weights[i] / det
        coeffs[i] = coeff

        # Casertano+22 Eq. 38
        ramp_fit.slope += coeff * resultants[i]

        # Casertano+22 Eq. 39
        ramp_fit.read_var += (coeff ** 2 * read_noise ** 2 / n_reads[i])

        # Casertano+22 Eq 40
        #    Note that this is an inversion of the indexing from the equation;
        #    however, commutivity of addition results in the same answer. This
        #    makes it so that we don't have to loop over all the resultants twice.
        ramp_fit.poisson_var += coeff ** 2 * tau[i]
        for j in range(i):
            ramp_fit.poisson_var += (2 * coeff * coeffs[j] * t_bar[j])

    return ramp_fit



import logging

import numpy as np
from gwcs.wcstools import grid_from_bounding_box

from stcal.alignment import util

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def calc_pixmap(in_wcs, out_wcs, shape=None):
    """Return a pixel grid map from input frame to output frame.

    Parameters
    ----------
    in_wcs : `~astropy.wcs.WCS`
        Input WCS objects or transforms.
    out_wcs : `~astropy.wcs.WCS` or `~gwcs.wcs.WCS`
        output WCS objects or transforms.
    shape : tuple, optional
        Shape of grid in pixels. The default is None.

    Returns
    -------
    pixmap : ndarray of shape (xdim, ydim, 2)
        Reprojected pixel grid map. `pixmap[xin, yin]` returns `xout,
        yout` indices in the output image.
    """
    if shape:
        bb = util.wcs_bbox_from_shape(shape)
        log.debug("Bounding box from data shape: %s", bb)
    else:
        bb = util.wcs_bbox_from_shape(in_wcs.pixel_shape)
        log.debug("Bounding box from WCS: %s", bb)

    # creates 2 grids, one with rows of all x values * len(y) rows,
    # and the reverse for all y columns
    grid = grid_from_bounding_box(bb)
    transform_function = util.reproject(in_wcs, out_wcs)
    return np.dstack(transform_function(grid[0], grid[1]))



#! /usr/bin/env python
#
# utils.py: utility functions
import logging
import warnings

import numpy as np


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

# Replace zero or negative variances with this:
LARGE_VARIANCE = 1.0e8
LARGE_VARIANCE_THRESHOLD = 0.01 * LARGE_VARIANCE


class OptRes:
    """
    Object to hold optional results for all good pixels for
    y-intercept, slope, uncertainty for y-intercept, uncertainty for
    slope, inverse variance, first frame (for pedestal image), and
    cosmic ray magnitude.
    """

    def __init__(self, n_int, imshape, max_seg, nreads, save_opt):
        """
        Initialize the optional attributes. These are 4D arrays for the
        segment-specific values of the y-intercept, the slope, the uncertainty
        associated with both, the weights, the approximate cosmic ray
        magnitudes, and the inverse variance.  These are 3D arrays for the
        integration-specific first frame and pedestal values.

        Parameters
        ----------
        n_int : int
            number of integrations in data set

        imshape : tuple
            shape of 2D image

        max_seg : int
            maximum number of segments fit

        nreads : int
            number of reads in an integration

        save_opt : bool
           save optional fitting results
        """
        self.slope_seg = np.zeros((n_int, max_seg, *imshape), dtype=np.float32)
        if save_opt:
            self.yint_seg = np.zeros((n_int, max_seg, *imshape), dtype=np.float32)
            self.sigyint_seg = np.zeros((n_int, max_seg, *imshape), dtype=np.float32)
            self.sigslope_seg = np.zeros((n_int, max_seg, *imshape), dtype=np.float32)
            self.inv_var_seg = np.zeros((n_int, max_seg, *imshape), dtype=np.float32)
            self.firstf_int = np.zeros((n_int, *imshape), dtype=np.float32)
            self.ped_int = np.zeros((n_int, *imshape), dtype=np.float32)
            self.cr_mag_seg = np.zeros((n_int, nreads, *imshape), dtype=np.float32)
            self.var_p_seg = np.zeros((n_int, max_seg, *imshape), dtype=np.float32)
            self.var_r_seg = np.zeros((n_int, max_seg, *imshape), dtype=np.float32)

    def init_2d(self, npix, max_seg, save_opt):
        """
        Initialize the 2D segment-specific attributes for the current data
        section.

        Parameters
        ----------
        npix : integer
            number of pixels in section of 2D array

        max_seg : integer
            maximum number of segments that will be fit within an
            integration, calculated over all pixels and all integrations

        save_opt : bool
            save optional fitting results

        Returns
        -------
        None
        """
        self.slope_2d = np.zeros((max_seg, npix), dtype=np.float32)
        if save_opt:
            self.interc_2d = np.zeros((max_seg, npix), dtype=np.float32)
            self.siginterc_2d = np.zeros((max_seg, npix), dtype=np.float32)
            self.sigslope_2d = np.zeros((max_seg, npix), dtype=np.float32)
            self.inv_var_2d = np.zeros((max_seg, npix), dtype=np.float32)
            self.firstf_2d = np.zeros((max_seg, npix), dtype=np.float32)
            self.var_s_2d = np.zeros((max_seg, npix), dtype=np.float32)
            self.var_r_2d = np.zeros((max_seg, npix), dtype=np.float32)

    def reshape_res(self, num_int, rlo, rhi, sect_shape, ff_sect, save_opt):
        """
        Loop over the segments and copy the reshaped 2D segment-specific
        results for the current data section to the 4D output arrays.

        Parameters
        ----------
        num_int : int
            integration number

        rlo : int
            first column of section

        rhi : int
            last column of section

        sect_sect : tuple
            shape of section image

        ff_sect : ndarray
            first frame data, 2-D float

        save_opt : bool
            save optional fitting results
        """
        for ii_seg in range(self.slope_seg.shape[1]):
            self.slope_seg[num_int, ii_seg, rlo:rhi, :] = self.slope_2d[ii_seg, :].reshape(sect_shape)

            if save_opt:
                self.yint_seg[num_int, ii_seg, rlo:rhi, :] = self.interc_2d[ii_seg, :].reshape(sect_shape)
                self.slope_seg[num_int, ii_seg, rlo:rhi, :] = self.slope_2d[ii_seg, :].reshape(sect_shape)
                self.sigyint_seg[num_int, ii_seg, rlo:rhi, :] = self.siginterc_2d[ii_seg, :].reshape(
                    sect_shape
                )
                self.sigslope_seg[num_int, ii_seg, rlo:rhi, :] = self.sigslope_2d[ii_seg, :].reshape(
                    sect_shape
                )
                self.inv_var_seg[num_int, ii_seg, rlo:rhi, :] = self.inv_var_2d[ii_seg, :].reshape(sect_shape)
                self.firstf_int[num_int, rlo:rhi, :] = ff_sect

    def append_arr(self, num_seg, g_pix, intercept, slope, sig_intercept, sig_slope, inv_var, save_opt):
        """
        Add the fitting results for the current segment to the 2d arrays.

        Parameters
        ----------
        num_seg : ndarray
            counter for segment number within the section, 1-D int

        g_pix : ndarray
            pixels having fitting results in current section, 1-D int

        intercept : ndarray
            intercepts for pixels in current segment and section, 1-D float

        slope : ndarray
            slopes for pixels in current segment and section, 1-D float

        sig_intercept : ndarray
            uncertainties of intercepts for pixels in current segment
            and section, 1-D float

        sig_slope : ndarray
            uncertainties of slopes for pixels in current segment and
            section, 1-D float

        inv_var : ndarray
            reciprocals of variances for fits of pixels in current
            segment and section, 1-D float

        save_opt : bool
            save optional fitting results

        Returns
        -------
        None
        """
        '''
        if False:
            print("=" * 80)
            dbg_print(f"slope         = {slope}")
            dbg_print(f"intercept     = {intercept}")
            dbg_print(f"inv_var       = {inv_var}")
            dbg_print(f"sig_intercept = {sig_intercept}")
            dbg_print(f"sig_slope     = {sig_slope}")
            print("=" * 80)
        '''

        self.slope_2d[num_seg[g_pix], g_pix] = slope[g_pix]

        if save_opt:
            self.interc_2d[num_seg[g_pix], g_pix] = intercept[g_pix]
            self.siginterc_2d[num_seg[g_pix], g_pix] = sig_intercept[g_pix]
            self.sigslope_2d[num_seg[g_pix], g_pix] = sig_slope[g_pix]
            self.inv_var_2d[num_seg[g_pix], g_pix] = inv_var[g_pix]

    def shrink_crmag(self, n_int, dq_cube, imshape, nreads, jump_det):
        """
        Compress the 4D cosmic ray magnitude array for the current
        integration, removing all groups whose cr magnitude is 0 for
        pixels having at least one group with a non-zero magnitude. For
        every integration, the depth of the array is equal to the
        maximum number of cosmic rays flagged in all pixels in all
        integrations.  This routine currently involves a loop over all
        pixels having at least 1 group flagged; if this algorithm takes
        too long for datasets having an overabundance of cosmic rays,
        this routine will require further optimization.

        Parameters
        ----------
        n_int : int
            number of integrations in dataset

        dq_cube : ndarray
            input data quality array, 4-D flag

        imshape : tuple
            shape of a single input image

        nreads : int
            number of reads in an integration

        Returns
        -------
        None

        """
        # Loop over data integrations to find max num of crs flagged per pixel
        # (this could exceed the maximum number of segments fit)
        max_cr = 0
        for ii_int in range(n_int):
            dq_int = dq_cube[ii_int, :, :, :]
            dq_cr = np.bitwise_and(jump_det, dq_int)
            max_cr_int = (dq_cr > 0.0).sum(axis=0).max()
            max_cr = max(max_cr, max_cr_int)

        # Allocate compressed array based on max number of crs
        cr_com = np.zeros((n_int, max_cr, *imshape), dtype=np.float32)

        # Loop over integrations and groups: for those pix having a cr, add
        #    the magnitude to the compressed array
        for ii_int in range(n_int):
            cr_mag_int = self.cr_mag_seg[ii_int, :, :, :]
            cr_int_has_cr = np.where(cr_mag_int.sum(axis=0) != 0)

            # Initialize number of crs for each image pixel for this integration
            end_cr = np.zeros(imshape, dtype=np.int16)

            for k_rd in range(nreads):
                # loop over pixels having a CR
                for nn in range(len(cr_int_has_cr[0])):
                    y, x = cr_int_has_cr[0][nn], cr_int_has_cr[1][nn]

                    if cr_mag_int[k_rd, y, x] > 0.0:
                        cr_com[ii_int, end_cr[y, x], y, x] = cr_mag_int[k_rd, y, x]
                        end_cr[y, x] += 1

        max_num_crs = end_cr.max()
        if max_num_crs == 0:
            max_num_crs = 1
            self.cr_mag_seg = np.zeros(shape=(n_int, 1, imshape[0], imshape[1]))
        else:
            self.cr_mag_seg = cr_com[:, :max_num_crs, :, :]

    def output_optional(self, group_time):
        """
        These results are the cosmic ray magnitudes in the
        segment-specific results for the count rates, y-intercept,
        uncertainty in the slope, uncertainty in the y-intercept,
        pedestal image, fitting weights, and the uncertainties in
        the slope due to poisson noise only and read noise only, and
        the integration-specific results for the pedestal image.  The
        slopes are divided by the effective integration time here to
        yield the count rates. Any variance values that are a large fraction
        of the default value LARGE_VARIANCE correspond to non-existent segments,
        so will be set to 0 here before output.

        Parameters
        ----------
        group_time : float
            effective integration time for a single group

        Returns
        -------
        opt_info : tuple
            The tuple of computed optional results arrays for fitting.
        """
        self.var_p_seg[self.var_p_seg > LARGE_VARIANCE_THRESHOLD] = 0.0
        self.var_r_seg[self.var_r_seg > LARGE_VARIANCE_THRESHOLD] = 0.0

        # Suppress, then re-enable, arithmetic warnings
        warnings.filterwarnings("ignore", ".*invalid value.*", RuntimeWarning)
        warnings.filterwarnings("ignore", ".*divide by zero.*", RuntimeWarning)

        # Tiny 'weights' values correspond to non-existent segments, so set to 0.
        self.weights[1.0 / self.weights > LARGE_VARIANCE_THRESHOLD] = 0.0
        warnings.resetwarnings()

        return (
            self.slope_seg,
            self.sigslope_seg,
            self.var_p_seg,
            self.var_r_seg,
            self.yint_seg,
            self.sigyint_seg,
            self.ped_int,
            self.weights,
            self.cr_mag_seg,
        )

    def print_full(self):  # pragma: no cover
        """
        Diagnostic function for printing optional output arrays; most
        useful for tiny datasets.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        print("Will now print all optional output arrays - ")
        print(" yint_seg: ")
        print(self.yint_seg)
        print("  ")
        print(" slope_seg: ")
        print(self.slope_seg)
        print("  ")
        print(" sigyint_seg: ")
        print(self.sigyint_seg)
        print("  ")
        print(" sigslope_seg: ")
        print(self.sigslope_seg)
        print("  ")
        print(" inv_var_2d: ")
        print(self.inv_var_2d)
        print("  ")
        print(" firstf_int: ")
        print(self.firstf_int)
        print("  ")
        print(" ped_int: ")
        print(self.ped_int)
        print("  ")
        print(" cr_mag_seg: ")
        print(self.cr_mag_seg)


def alloc_arrays_1(n_int, imshape):
    """
    Allocate arrays for integration-specific results and segment-specific
    results and variances.

    Parameters
    ----------
    n_int : int
        number of integrations

    imshape : tuple
        shape of a single image

    Returns
    -------
    dq_int : ndarray
        Cube of integration-specific group data quality values, 3-D flag

    median_diffs_2d : ndarray
        Estimated median slopes, 2-D float

    num_seg_per_int : ndarray
        Cube of numbers of segments for all integrations and pixels, 3-D int

    sat_0th_group_int : ndarray
        Integration-specific slice whose value for a pixel is 1 if the initial
        group of the ramp is saturated, 3-D uint8
    """
    dq_int = np.zeros((n_int, *imshape), dtype=np.uint32)
    num_seg_per_int = np.zeros((n_int, *imshape), dtype=np.uint8)

    # for estimated median slopes
    sat_0th_group_int = np.zeros((n_int, *imshape), dtype=np.uint8)

    return dq_int, num_seg_per_int, sat_0th_group_int


def alloc_arrays_2(n_int, imshape, max_seg):
    """
    Allocate arrays for integration-specific results and segment-specific
    results and variances.

    Parameters
    ----------
    n_int : int
        number of integrations

    imshape : tuple
        shape of a single image

    max_seg : int
        maximum number of segments fit

    Returns
    -------
    var_p3 : ndarray
        Cube of integration-specific values for the slope variance due to
        Poisson noise only, 3-D float

    var_r3 : ndarray
        Cube of integration-specific values for the slope variance due to
        readnoise only, 3-D float

    var_p4 : ndarray
        Hypercube of segment- and integration-specific values for the slope
        variance due to Poisson noise only, 4-D float

    var_r4 : ndarray
        Hypercube of segment- and integration-specific values for the slope
        variance due to read noise only, 4-D float

    var_both4 : ndarray
        Hypercube of segment- and integration-specific values for the slope
        variance due to combined Poisson noise and read noise, 4-D float

    var_both3 : ndarray
        Cube of segment- and integration-specific values for the slope
        variance due to combined Poisson noise and read noise, 3-D float

    inv_var_both4 : ndarray
        Hypercube of reciprocals of segment- and integration-specific values for
        the slope variance due to combined Poisson noise and read noise, 4-D float

    s_inv_var_p3 : ndarray
        Cube of reciprocals of segment- and integration-specific values for
        the slope variance due to Poisson noise only, summed over segments, 3-D float

    s_inv_var_r3 : ndarray
        Cube of reciprocals of segment- and integration-specific values for
        the slope variance due to read noise only, summed over segments, 3-D float

    s_inv_var_both3 : ndarray
        Cube of reciprocals of segment- and integration-specific values for
        the slope variance due to combined Poisson noise and read noise, summed
        over segments, 3-D float

    segs_4 : ndarray
        Hypercube of lengths of segments for all integrations and pixels, 4-D
        int
    """
    # Initialize variances so that non-existing ramps and segments will have
    #   negligible contributions
    # Integration-specific:
    var_p3 = np.zeros((n_int, *imshape), dtype=np.float32) + LARGE_VARIANCE
    var_r3 = var_p3.copy()
    var_both3 = var_p3.copy()
    s_inv_var_p3 = np.zeros_like(var_p3)
    s_inv_var_r3 = np.zeros_like(var_p3)
    s_inv_var_both3 = np.zeros_like(var_p3)

    # Segment-specific:
    var_p4 = np.zeros((n_int, max_seg, *imshape), dtype=np.float32) + LARGE_VARIANCE
    var_r4 = var_p4.copy()
    var_both4 = var_p4.copy()
    inv_var_both4 = np.zeros_like(var_p4)

    # number of segments
    segs_4 = np.zeros((n_int, max_seg, *imshape), dtype=np.uint8)

    return (
        var_p3,
        var_r3,
        var_p4,
        var_r4,
        var_both4,
        var_both3,
        inv_var_both4,
        s_inv_var_p3,
        s_inv_var_r3,
        s_inv_var_both3,
        segs_4,
    )


def calc_slope_vars(ramp_data, rn_sect, gain_sect, gdq_sect, group_time, max_seg):
    """
    Calculate the segment-specific variance arrays for the given
    integration.

    Parameters
    ----------
    ramp_data : ramp_fit_class.RampData
        Contains the DQ flags needed for this function

    rn_sect : ndarray
        read noise values for all pixels in data section, 2-D float

    gain_sect : ndarray
        gain values for all pixels in data section, 2-D float

    gdq_sect : ndarray
        data quality flags for pixels in section, 3-D int

    group_time : float
        Time increment between groups, in seconds.

    max_seg : int
        maximum number of segments fit

    Returns
    -------
    den_r3 : ndarray
        for a given integration, the reciprocal of the denominator of the
        segment-specific variance of the segment's slope due to read noise, 3-D float

    den_p3 : ndarray
        for a given integration, the reciprocal of the denominator of the
        segment-specific variance of the segment's slope due to Poisson noise, 3-D float

    num_r3 : ndarray
        numerator of the segment-specific variance of the segment's slope
        due to read noise, 3-D float

    segs_beg_3 : ndarray
        lengths of segments for all pixels in the given data section and
        integration, 3-D int
    """
    (ngroups, nrows, ncols) = gdq_sect.shape
    npix = nrows * ncols
    imshape = (nrows, ncols)

    # Create integration-specific sections of input arrays for determination
    #   of the variances.
    gdq_2d = gdq_sect[:, :, :].reshape((ngroups, npix))
    gain_1d = gain_sect.reshape(npix)
    gdq_2d_nan = gdq_2d.copy()  # group dq with SATS will be replaced by nans
    gdq_2d_nan = gdq_2d_nan.astype(np.float32)

    # set all SAT groups to nan
    gdq_2d_nan[np.bitwise_and(gdq_2d, ramp_data.flags_saturated).astype(bool)] = np.nan

    # Get lengths of semiramps for all pix [number_of_semiramps, number_of_pix]
    segs = np.zeros_like(gdq_2d).astype(np.uint16)

    # Counter of semiramp for each pixel
    sr_index = np.zeros(npix, dtype=np.uint16)
    pix_not_done = np.ones(npix, dtype=bool)  # initialize to True

    group = 0
    # Loop over reads for all pixels to get segments (segments per pixel)
    while group < ngroups and np.any(pix_not_done):
        gdq_1d = gdq_2d_nan[group, :]
        wh_good = np.where(gdq_1d == 0)  # good groups

        # if this group is good, increment those pixels' segments' lengths
        if len(wh_good[0]) > 0:
            segs[sr_index[wh_good], wh_good] += 1
        del wh_good

        # Locate any CRs that appear before the first SAT group...
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "invalid value.*", RuntimeWarning)
            wh_cr = np.where(gdq_2d_nan[group, :].astype(np.int32) & ramp_data.flags_jump_det > 0)

        # ... but not on final read:
        if len(wh_cr[0]) > 0 and (group < ngroups - 1):
            sr_index[wh_cr[0]] += 1
            segs[sr_index[wh_cr], wh_cr] += 1

        del wh_cr

        # If current group is a NaN, this pixel is done (pix_not_done is False)
        wh_nan = np.where(np.isnan(gdq_2d_nan[group, :]))
        if len(wh_nan[0]) > 0:
            pix_not_done[wh_nan[0]] = False

        del wh_nan

        group += 1

    segs = segs.astype(np.uint16)
    segs_beg = segs[:max_seg, :]  # the leading nonzero lengths

    # Create reshaped version [ segs, y, x ] to simplify computation
    segs_beg_3 = segs_beg.reshape(max_seg, imshape[0], imshape[1])
    segs_beg_3 = remove_bad_singles(segs_beg_3)

    # Create a version 1 less for later calculations for the variance due to
    #   Poisson, with a floor=1 to handle single-group segments
    segs_beg_3_m1 = segs_beg_3.copy()
    segs_beg_3_m1[segs_beg_3 > 1] -= 1
    segs_beg_3_m1[segs_beg_3_m1 < 1] = 1

    # For a segment, the variance due to Poisson noise
    #   = slope/(tgroup * gain * (ngroups-1)),
    #   where slope is the estimated median slope, tgroup is the group time,
    #   and ngroups is the number of groups in the segment.
    #   Here the denominator of this quantity will be computed, which will be
    #   later multiplied by the estimated median slope.

    # Suppress, then re-enable, harmless arithmetic warnings, as NaN will be
    #   checked for and handled later
    warnings.filterwarnings("ignore", ".*invalid value.*", RuntimeWarning)
    warnings.filterwarnings("ignore", ".*divide by zero.*", RuntimeWarning)
    den_p3 = 1.0 / (group_time * gain_1d.reshape(imshape) * segs_beg_3_m1)

    if ramp_data.zframe_locs:
        zinteg_locs = ramp_data.zframe_locs[ramp_data.current_integ]
        frame_time = ramp_data.frame_time
        tmp_den_p3 = den_p3[0, :, :]
        tmp_den_p3[zinteg_locs] = 1.0 / (frame_time * gain_sect[zinteg_locs])
        den_p3[0, :, :] = tmp_den_p3

    if ramp_data.one_groups_time is not None:
        ginteg_locs = ramp_data.one_groups_locs[ramp_data.current_integ]
        tmp_den_p3 = den_p3[0, :, :]
        tmp_den_p3[ginteg_locs] = 1.0 / (ramp_data.one_groups_time * gain_sect[ginteg_locs])
        den_p3[0, :, :] = tmp_den_p3

    warnings.resetwarnings()

    # For a segment, the variance due to readnoise noise
    # = 12 * readnoise**2 /(ngroups_seg**3. - ngroups_seg)/( tgroup **2.)
    num_r3 = 12.0 * (rn_sect / group_time) ** 2.0  # always >0

    if ramp_data.zframe_locs:
        zinteg_locs = ramp_data.zframe_locs[ramp_data.current_integ]
        frame_time = ramp_data.frame_time
        num_r3[zinteg_locs] = 12.0 * (rn_sect[zinteg_locs] / frame_time) ** 2.0

    if ramp_data.one_groups_time is not None:
        ginteg_locs = ramp_data.one_groups_locs[ramp_data.current_integ]
        num_r3[ginteg_locs] = 12.0 * (rn_sect[ginteg_locs] / ramp_data.one_groups_time) ** 2.0

    # Reshape for every group, every pixel in section
    num_r3 = np.dstack([num_r3] * max_seg)
    num_r3 = np.transpose(num_r3, (2, 0, 1))

    # Denominator den_r3 = 1./(segs_beg_3 **3.-segs_beg_3). The minimum number
    #   of allowed groups is 2, which will apply if there is actually only 1
    #   group; in this case den_r3 = 1/6. This covers the case in which there is
    #   only one good group at the beginning of the integration, so it will be
    #   be compared to the plane of (near) zeros resulting from the reset. For
    #   longer segments, this value is overwritten below.
    den_r3 = num_r3.copy() * 0.0 + 1.0 / 6
    wh_seg_pos = np.where(segs_beg_3 > 1)

    # Suppress, then, re-enable harmless arithmetic warnings, as NaN will be
    #   checked for and handled later
    warnings.filterwarnings("ignore", ".*invalid value.*", RuntimeWarning)
    warnings.filterwarnings("ignore", ".*divide by zero.*", RuntimeWarning)
    # overwrite where segs>1
    den_r3[wh_seg_pos] = 1.0 / (segs_beg_3[wh_seg_pos] ** 3.0 - segs_beg_3[wh_seg_pos])
    warnings.resetwarnings()

    return den_r3, den_p3, num_r3, segs_beg_3


def calc_pedestal(ramp_data, num_int, slope_int, firstf_int, dq_first, nframes, groupgap, dropframes1):
    """
    The pedestal is calculated by extrapolating the final slope for each pixel
    from its value at the first sample in the integration to an exposure time
    of zero; this calculation accounts for the values of nframes and groupgap.
    Any pixel that is saturated on the 1st group is given a pedestal value of 0.

    Parameters
    ----------
    num_int : int
        integration number

    slope_int : ndarray
        cube of integration-specific slopes, 3-D float

    firstf_int : ndarray
        integration-specific first frame array, 3-D float

    dq_first : ndarray
        DQ of the initial group for all ramps in the given integration, 2-D
        flag

    nframes : int
        number of frames averaged per group; from the NFRAMES keyword. Does
        not contain the groupgap.

    groupgap : int
        number of frames dropped between groups, from the GROUPGAP keyword.

    dropframes1 : int
        number of frames dropped at the beginning of every integration, from
        the DRPFRMS1 keyword.

    Returns
    -------
    ped : ndarray
        pedestal image, 2-D float
    """
    ff_all = firstf_int[num_int, :, :].astype(np.float32)
    tmp = ((nframes + 1.0) / 2.0 + dropframes1) / (nframes + groupgap)
    ped = ff_all - slope_int[num_int, ::] * tmp

    sat_flag = ramp_data.flags_saturated
    ped[np.bitwise_and(dq_first, sat_flag) == sat_flag] = 0
    ped[np.isnan(ped)] = 0.0

    return ped


def output_integ(ramp_data, slope_int, dq_int, var_p3, var_r3, var_both3):
    """
    For the OLS algorithm, construct the output integration-specific results.
    Any variance values that are a large fraction of the default value
    LARGE_VARIANCE correspond to non-existent segments, so will be set to 0
    here before output.

    Parameters
    ----------
    ramp_data : RampData
        Contains flag information.

    model : instance of Data Model
       DM object for input

    slope_int : ndarray
       Data cube of weighted slopes for each integration, 3-D float

    dq_int : ndarray
       Data cube of DQ arrays for each integration, 3-D int

    var_p3 : ndarray
        Cube of integration-specific values for the slope variance due to
        Poisson noise only, 3-D float

    var_r3 : ndarray
        Cube of integration-specific values for the slope variance due to
        read noise only, 3-D float

    var_both3 : ndarray
        Cube of integration-specific values for the slope variance due to
        read noise and Poisson noise, 3-D float

    Returns
    -------
    integ_info : tuple
        The tuple of computed integration ramp fitting arrays.

    """
    # Suppress harmless arithmetic warnings for now
    warnings.filterwarnings("ignore", ".*invalid value.*", RuntimeWarning)
    warnings.filterwarnings("ignore", ".*divide by zero.*", RuntimeWarning)

    var_p3[var_p3 > LARGE_VARIANCE_THRESHOLD] = 0.0
    var_r3[var_r3 > LARGE_VARIANCE_THRESHOLD] = 0.0
    var_both3[var_both3 > LARGE_VARIANCE_THRESHOLD] = 0.0

    data = slope_int
    invalid_data = ramp_data.flags_saturated | ramp_data.flags_do_not_use
    data[np.bitwise_and(dq_int, invalid_data).astype(bool)] = np.nan

    err = np.sqrt(var_both3)
    dq = dq_int
    var_poisson = var_p3

    var_rnoise = var_r3
    integ_info = (data, dq, var_poisson, var_rnoise, err)

    # Reset the warnings filter to its original state
    warnings.resetwarnings()

    return integ_info


def gls_pedestal(first_group, slope_int, s_mask, frame_time, nframes_used):  # pragma: no cover
    """
    Calculate the pedestal for the GLS case.

    The pedestal is the first group, but extrapolated back to zero time
    using the slope obtained by the fit to the whole ramp.  The time of
    the first group is the frame time multiplied by (M + 1) / 2, where M
    is the number of frames per group, not including the number (if any)
    of skipped frames.

    The input arrays and output pedestal are slices of the full arrays.
    They are just the relevant data for the current integration (assuming
    that this function is called within a loop over integrations), and
    they may include only a subset of image lines.  For example, this
    function might be called with slope_int argument given as:
    `slope_int[num_int, rlo:rhi, :]`.  The input and output parameters
    are in electrons.

    Parameters
    ----------
    first_group : ndarray
        A slice of the first group in the ramp, 2-D float

    slope_int : ndarray
        The slope obtained by GLS ramp fitting.  This is a slice for the
        current integration and a subset of the image lines, 2-D float

    s_mask : ndarray
        True for ramps that were saturated in the first group, 2-D bool

    frame_time : float
        The time to read one frame, in seconds.

    nframes_used : int
        Number of frames that were averaged together to make a group.
        Excludes the groupgap.

    Returns
    -------
    pedestal : ndarray
        This is a slice of the full pedestal array, and it's for the
        current integration, 2-D float
    """
    M = float(nframes_used)
    pedestal = first_group - slope_int * frame_time * (M + 1.0) / 2.0
    if s_mask.any():
        pedestal[s_mask] = 0.0

    return pedestal


def shift_z(a, off):
    """
    Shift input 3D array by requested offset in z-direction, padding shifted
    array (of the same size) by leading or trailing zeros as needed.

    Parameters
    ----------
    a : ndarray
        input array, 3-D float

    off : int
        offset in z-direction

    Returns
    -------
    b : ndarray
        shifted array, 3-D float

    """
    # set initial and final indices along z-direction for original and
    #    shifted 3D arrays
    ai_z = int((abs(off) + off) / 2)
    af_z = a.shape[0] + int((-abs(off) + off) / 2)

    bi_z = a.shape[0] - af_z
    bf_z = a.shape[0] - ai_z

    b = a * 0
    b[bi_z:bf_z, :, :] = a[ai_z:af_z, :, :]

    return b


def get_efftim_ped(ramp_data):
    """
    XXX - Work to remove this function.
    Calculate the effective integration time for a single group, and return the
    number of frames per group, and the number of frames dropped between groups.

    Parameters
    ----------
    ramp_data: RampClass
        Object for input data.

    Returns
    -------
    effintim : float
        effective integration time for a single group

    nframes : int
        number of frames averaged per group; from the NFRAMES keyword.

    groupgap : int
        number of frames dropped between groups; from the GROUPGAP keyword.

    dropframes1 : int
        number of frames dropped at the beginning of every integration; from
        the DRPFRMS1 keyword, or 0 if the keyword is missing
    """
    groupgap = ramp_data.groupgap
    nframes = ramp_data.nframes
    frame_time = ramp_data.frame_time
    dropframes1 = ramp_data.drop_frames1

    if dropframes1 is None:  # set to default if missing
        dropframes1 = 0
        log.debug("Missing keyword DRPFRMS1, so setting to default value of 0")

    try:
        effintim = (nframes + groupgap) * frame_time
    except TypeError:
        log.exception("Can not retrieve values needed to calculate integ. time")

    log.debug("Calculating effective integration time for a single group using:")
    log.debug(" groupgap: %s", groupgap)
    log.debug(" nframes: %s", nframes)
    log.debug(" frame_time: %s", frame_time)
    log.debug(" dropframes1: %s", dropframes1)
    log.info("Effective integration time per group: %s", effintim)

    return effintim, nframes, groupgap, dropframes1


def get_dataset_info(ramp_data):
    """
    Extract values for the number of groups, the number of pixels, dataset
    shapes, the number of integrations, the instrument name, the frame time,
    and the observation time.

    Parameters
    ----------
    ramp_data: RampClass
       Object for input data.

    Returns
    -------
    nreads : int
       number of reads in input dataset

    npix : int
       number of pixels in 2D array

    imshape : tuple
       shape of 2D image

    cubeshape : tuple
       shape of input dataset

    n_int : int
       number of integrations

    instrume : str
       instrument

    frame_time : float
       integration time from TGROUP

    ngroups : int
        number of groups per integration

    group_time : float
        Time increment between groups, in seconds.
    """
    instrume = ramp_data.instrument_name
    frame_time = ramp_data.frame_time
    ngroups = ramp_data.data.shape[1]
    group_time = ramp_data.group_time

    n_int = ramp_data.data.shape[0]
    nreads = ramp_data.data.shape[1]
    asize2 = ramp_data.data.shape[2]
    asize1 = ramp_data.data.shape[3]

    npix = asize2 * asize1  # number of pixels in 2D array
    imshape = (asize2, asize1)
    cubeshape = (nreads, *imshape)

    return (nreads, npix, imshape, cubeshape, n_int, instrume, frame_time, ngroups, group_time)


def get_more_info(ramp_data, saturated_flag, jump_flag):  # pragma: no cover
    """
    Get information used by GLS algorithm.

    Parameters
    ----------
    ramp_data: RampClass
        Object for input data.

    Returns
    -------
    group_time : float
        Time increment between groups, in seconds.

    nframes_used : int
        Number of frames that were averaged together to make a group,
        i.e. excluding skipped frames.

    saturated_flag : int
        Group data quality flag that indicates a saturated pixel.

    jump_flag : int
        Group data quality flag that indicates a cosmic ray hit.
    """
    group_time = ramp_data.group_time
    nframes_used = ramp_data.nframes
    saturated_flag = ramp_data.flags_saturated
    jump_flag = ramp_data.flags_jump_det

    return group_time, nframes_used, saturated_flag, jump_flag


def get_max_num_cr(gdq_cube, jump_flag):  # pragma: no cover
    """
    Find the maximum number of cosmic-ray hits in any one pixel.

    Parameters
    ----------
    gdq_cube : ndarray
        The group data quality array, 3-D flag

    jump_flag : int
        The data quality flag indicating a cosmic-ray hit.

    Returns
    -------
    max_num_cr : int
        The maximum number of cosmic-ray hits for any pixel.
    """
    cr_flagged = np.empty(gdq_cube.shape, dtype=np.uint8)
    cr_flagged[:] = np.where(np.bitwise_and(gdq_cube, jump_flag), 1, 0)
    max_num_cr = cr_flagged.sum(axis=0, dtype=np.int32).max()
    del cr_flagged

    return max_num_cr


def reset_bad_gain(ramp_data, pdq, gain):
    """
    For pixels in the gain array that are either non-positive or NaN, reset the
    the corresponding pixels in the pixel DQ array to NO_GAIN_VALUE and
    DO_NOT_USE so that they will be ignored.

    Parameters
    ----------
    pdq : ndarray
        pixel dq array of input model, 2-D int

    gain : ndarray
        gain array from reference file, 2-D float

    Returns
    -------
    pdq : ndarray
        pixleldq array of input model, reset to NO_GAIN_VALUE and DO_NOT_USE
        for pixels in the gain array that are either non-positive or NaN., 2-D
        flag
    """
    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "invalid value.*", RuntimeWarning)
        wh_g = np.where(gain <= 0.)
    """
    wh_g = np.where(gain <= 0.0)
    if len(wh_g[0]) > 0:
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], ramp_data.flags_no_gain_val)
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], ramp_data.flags_do_not_use)

    wh_g = np.where(np.isnan(gain))
    if len(wh_g[0]) > 0:
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], ramp_data.flags_no_gain_val)
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], ramp_data.flags_do_not_use)

    return pdq


def remove_bad_singles(segs_beg_3):
    """
    For the current integration and data section, remove all segments having only
    a single group if there are other segments in the ramp.  This method allows
    for the possibility that a ramp can have multiple (necessarily consecutive)
    1-group segments, which in principle could occur if there are consecutive
    cosmic rays.

    Parameters
    ----------
    segs_beg_3 : ndarray
        lengths of all segments for all ramps in the given data section and
        integration; some of these ramps may contain segments having a single
        group, and another segment, 3-D int

    Returns
    -------
    segs_beg_3 : ndarray
        lengths of all segments for all ramps in the given data section and
        integration; segments having a single group, and another segment
        will be removed, 3-D int
    """
    max_seg = segs_beg_3.shape[0]

    # get initial number of ramps having single-group segments
    tot_num_single_grp_ramps = len(np.where((segs_beg_3 == 1) & (segs_beg_3.sum(axis=0) > 1))[0])

    while tot_num_single_grp_ramps > 0:
        # until there are no more single-group segments
        for ii_0 in range(max_seg):
            slice_0 = segs_beg_3[ii_0, :, :]

            for ii_1 in range(max_seg):  # correctly includes EARLIER segments
                if ii_0 == ii_1:  # don't compare with itself
                    continue

                slice_1 = segs_beg_3[ii_1, :, :]

                # Find ramps of a single-group segment and another segment
                # either earlier or later
                wh_y, wh_x = np.where((slice_0 == 1) & (slice_1 > 0))

                if len(wh_y) == 0:
                    # Are none, so go to next pair of segments to check
                    continue

                # Remove the 1-group segment
                segs_beg_3[ii_0:-1, wh_y, wh_x] = segs_beg_3[ii_0 + 1 :, wh_y, wh_x]

                # Zero the last segment entry for the ramp, which would otherwise
                # remain non-zero due to the shift
                segs_beg_3[-1, wh_y, wh_x] = 0

                del wh_y, wh_x

                tot_num_single_grp_ramps = len(np.where((segs_beg_3 == 1) & (segs_beg_3.sum(axis=0) > 1))[0])

    return segs_beg_3


def fix_sat_ramps(ramp_data, sat_0th_group_int, var_p3, var_both3, slope_int, dq_int):
    """
    For ramps within an integration that are saturated on the initial group,
    reset the integration-specific variances and slope so they will have no
    contribution.

    Parameters
    ----------
    sat_0th_group_int : ndarray
        Integration-specific slice whose value for a pixel is 1 if the initial
        group of the ramp is saturated, 3-D uint8

    var_p3 : ndarray
        Cube of integration-specific values for the slope variance due to
        Poisson noise only; some ramps may be saturated in the initial group,
        3-D float

    var_both3 : ndarray
        Cube of segment- and integration-specific values for the slope
        variance due to combined Poisson noise and read noise; some ramps may
        be saturated in the initial group, 3-D float

    slope_int : ndarray
        Cube of integration-specific slopes. Some ramps may be saturated in the
        initial group, 3-D float

    dq_int : ndarray
        Cube of integration-specific DQ flags, 3-D flag

    Returns
    -------
    var_p3 : ndarray
        Cube of integration-specific values for the slope variance due to
        Poisson noise only; for ramps that are saturated in the initial group,
        this variance has been reset to a huge value to minimize the ramps
        contribution, 3-D float

    var_both3 : ndarray
        Cube of segment- and integration-specific values for the slope
        variance due to combined Poisson noise and read noise; for ramps that
        are saturated in the initial group, this variance has been reset to a
        huge value to minimize the ramps contribution, 3-D float

    slope_int : ndarray
        Cube of integration-specific slopes; for ramps that are saturated in
        the initial group, this variance has been reset to a huge value to
        minimize the ramps contribution, 3-D float

    dq_int : ndarray
        Cube of integration-specific DQ flags. For ramps that are saturated in
        the initial group, the flag 'DO_NOT_USE' is added, 3-D flag
    """
    var_p3[sat_0th_group_int > 0] = LARGE_VARIANCE
    var_both3[sat_0th_group_int > 0] = LARGE_VARIANCE
    slope_int[sat_0th_group_int > 0] = np.nan
    dq_int[sat_0th_group_int > 0] = np.bitwise_or(dq_int[sat_0th_group_int > 0], ramp_data.flags_do_not_use)

    return var_p3, var_both3, slope_int, dq_int


def do_all_sat(ramp_data, pixeldq, groupdq, imshape, n_int, save_opt):
    """
    For an input exposure where all groups in all integrations are saturated,
    the DQ in the primary and integration-specific output products are updated,
    and the other arrays in all output products are populated with zeros.

    Parameters
    ----------
    model : instance of Data Model
       DM object for input

    imshape : (int, int) tuple
       shape of 2D image

    n_int : int
       number of integrations

    save_opt : bool
       save optional fitting results

    Returns
    -------
    image_info : tuple
        The tuple of computed ramp fitting arrays.

    integ_info : tuple
        The tuple of computed integration fitting arrays.

    opt_info : tuple
        The tuple of computed optional results arrays for fitting.
    """
    # Create model for the primary output. Flag all pixels in the pixiel DQ
    #   extension as SATURATED and DO_NOT_USE.
    pixeldq = np.bitwise_or(pixeldq, ramp_data.flags_saturated)
    pixeldq = np.bitwise_or(pixeldq, ramp_data.flags_do_not_use)

    data = np.zeros(imshape, dtype=np.float32)
    dq = pixeldq
    var_poisson = np.zeros(imshape, dtype=np.float32)
    var_rnoise = np.zeros(imshape, dtype=np.float32)
    err = np.zeros(imshape, dtype=np.float32)
    image_info = (data, dq, var_poisson, var_rnoise, err)

    # Create model for the integration-specific output. The 3D group DQ created
    #   is based on the 4D group DQ of the model, and all pixels in all
    #   integrations will be flagged here as DO_NOT_USE (they are already flagged
    #   as SATURATED). The INT_TIMES extension will be left as None.
    if n_int > 1:
        m_sh = groupdq.shape  # (integ, grps/integ, y, x )
        groupdq_3d = np.zeros((m_sh[0], m_sh[2], m_sh[3]), dtype=np.uint32)

        for ii in range(n_int):  # add SAT flag to existing groupdq in each slice
            groupdq_3d[ii, :, :] = np.bitwise_or.reduce(groupdq[ii, :, :, :], axis=0)

        groupdq_3d = np.bitwise_or(groupdq_3d, ramp_data.flags_do_not_use)

        data = np.zeros((n_int, *imshape), dtype=np.float32)
        dq = groupdq_3d
        var_poisson = np.zeros((n_int, *imshape), dtype=np.float32)
        var_rnoise = np.zeros((n_int, *imshape), dtype=np.float32)
        err = np.zeros((n_int, *imshape), dtype=np.float32)

        integ_info = (data, dq, var_poisson, var_rnoise, err)
    else:
        integ_info = None

    # Create model for the optional output
    if save_opt:
        new_arr = np.zeros((n_int, 1, *imshape), dtype=np.float32)

        slope = new_arr
        sigslope = new_arr
        var_poisson = new_arr
        var_rnoise = new_arr
        yint = new_arr
        sigyint = new_arr
        pedestal = np.zeros((n_int, *imshape), dtype=np.float32)
        weights = new_arr
        crmag = new_arr

        opt_info = (slope, sigslope, var_poisson, var_rnoise, yint, sigyint, pedestal, weights, crmag)

    else:
        opt_info = None

    log.info("All groups of all integrations are saturated.")

    return image_info, integ_info, opt_info


def log_stats(c_rates):
    """
    Optionally log statistics of detected cosmic rays.

    Parameters
    ----------
    c_rates : ndarray
       weighted count rate, 2-D float

    Returns
    -------
    None
    """
    wh_c_0 = np.where(c_rates == 0.0)  # insuff data or no signal

    log.debug("The number of pixels having insufficient data")
    log.debug("due to excessive CRs or saturation %d:", len(wh_c_0[0]))
    log.debug(
        "Count rates - min, mean, max, std: %f, %f, %f, %f",
        c_rates.min(),
        c_rates.mean(),
        c_rates.max(),
        c_rates.std(),
    )


def compute_num_slices(max_cores, nrows, max_available):
    """
    Computes the number of slices to be created for multiprocessing.

    Parameters
    ----------
    max_cores : str
        Number of cores to use for multiprocessing. If set to 'none' (the default),
        then no multiprocessing will be done. The other allowable values are 'quarter',
        'half', and 'all' and string integers. This is the fraction of cores
        to use for multi-proc.
    nrows : int
        The number of rows that will be used across all process. This is the
        maximum number of slices to make sure that each process has some data.
    max_available: int
        This is the total number of cores available. The total number of cores
        includes the SMT cores (Hyper Threading for Intel).

    Returns
    -------
    number_slices : int
        The number of slices for multiprocessing.
    """
    number_slices = 1
    if max_cores.isnumeric():
        number_slices = int(max_cores)
    elif max_cores.lower() == "none" or max_cores.lower() == "one":
        number_slices = 1
    elif max_cores == "quarter":
        number_slices = max_available // 4 or 1
    elif max_cores == "half":
        number_slices = max_available // 2 or 1
    elif max_cores == "all":
        number_slices = max_available
    # Make sure we don't have more slices than rows or available cores.
    return min([nrows, number_slices, max_available])


def dq_compress_final(dq_int, ramp_data):
    """
    From the integration level DQ flags, set the final pixel DQ flags.

    Parameters
    ----------
    dq_int : ndarray
        The integration level DQ flags, 3-D (nints, nrows, ncols).

    ramp_data : RampData
        Contains the DQ flag information.

    Return
    ------
    final_dq : ndarray
        The final 2-D (nrows, ncols) pixel DQ array.
    """
    final_dq = dq_int[0, :, :]
    nints = dq_int.shape[0]
    for integ in range(1, nints):
        final_dq = np.bitwise_or(final_dq, dq_int[integ, :, :])

    dnu = np.uint32(ramp_data.flags_do_not_use)
    sat = np.uint32(ramp_data.flags_saturated)

    # Remove DO_NOT_USE and SATURATED because they need special handling.
    # These flags are not set in the final pixel DQ array by simply being set
    # in one of the integrations.
    not_sat_or_dnu = np.uint32(~(dnu | sat))
    final_dq = np.bitwise_and(final_dq, not_sat_or_dnu)

    # If all integrations are DO_NOT_USE or SATURATED, then set DO_NOT_USE.
    set_if_total_integ(final_dq, dq_int, dnu | sat, dnu)

    # If all integrations have SATURATED, then set DO_NOT_USE and SATURATED.
    set_if_total_integ(final_dq, dq_int, sat, dnu | sat)

    return final_dq


def set_if_total_integ(final_dq, integ_dq, flag, set_flag):
    """
    Set set_flag in final_dq if flag is present in all integrations.

    Parameters
    ----------
    final_dq : ndarray
        2-D array (nrows, ncols) of the final pixel DQ.

    integ_dq : ndarray
        3-D array (nints, nrows, ncols) of the integration level DQ.

    flag : int
        Flag to check in each integration.

    set_flag : int
        Flag to set if flag is found in each integration.
    """
    nints = integ_dq.shape[0]

    # Find where flag is set
    test_dq = np.zeros(integ_dq.shape, dtype=np.uint32)
    test_dq[np.bitwise_and(integ_dq, flag).astype(bool)] = 1

    # Sum over all integrations
    test_sum = test_dq.sum(axis=0)
    all_set = np.where(test_sum == nints)

    # If flag is set in all integrations, then set the set_flag
    final_dq[all_set] = np.bitwise_or(final_dq[all_set], set_flag)


def dq_compress_sect(ramp_data, num_int, gdq_sect, pixeldq_sect):
    """
    This sets the integration level flags for DO_NOT_USE, JUMP_DET and
    SATURATED.  If any ramp has a jump, this flag will be set for the
    integration.  If all groups in a ramp are flagged as DO_NOT_USE, then the
    integration level DO_NOT_USE flag will be set.  If a ramp is saturated in
    group 0, then the integration level flag is marked as SATURATED.  Also, if
    all groups are marked as DO_NOT_USE or SATURATED (as in suppressed one
    groups), then the DO_NOT_USE flag is set.

    Parameters
    ----------
    ramp_data : RampData
        Contains the DQ flag information.

    num_int : int
        The current integration number.

    gdq_sect : ndarray
        The current 3-D (ngroups, nrows, ncols) integration DQ array.

    pixeldq_sect : ndarray
        The 2-D (nrows, ncols) pixel DQ flags for the current integration.

    Return
    ------
    pixeldq_sect : ndarray
        The 2-D (nrows, ncols) pixel DQ flags for the current integration.
    """
    sat = ramp_data.flags_saturated
    jump = ramp_data.flags_jump_det
    dnu = ramp_data.flags_do_not_use
    ngroups, nrows, ncols = gdq_sect.shape

    # Check total SATURATED or DO_NOT_USE
    set_if_total_ramp(pixeldq_sect, gdq_sect, sat | dnu, dnu)

    # Assume total saturation if group 0 is SATURATED.
    gdq0_sat = np.bitwise_and(gdq_sect[0], sat)
    pixeldq_sect[gdq0_sat != 0] = np.bitwise_or(pixeldq_sect[gdq0_sat != 0], sat | dnu)

    # If jump occurs mark the appropriate flag.
    jump_loc = np.bitwise_and(gdq_sect, jump)
    jump_check = np.where(jump_loc.sum(axis=0) > 0)
    pixeldq_sect[jump_check] = np.bitwise_or(pixeldq_sect[jump_check], jump)

    return pixeldq_sect


def set_if_total_ramp(pixeldq_sect, gdq_sect, flag, set_flag):
    """
    Set set_flag in final_dq if flag is present in all integrations.

    Parameters
    ----------
    pixeldq_sect: ndarray
        2-D array (nrows, ncols) of the integration DQ.

    gdq_dq : ndarray
        3-D array (ngroups, nrows, ncols) of the integration level DQ.

    flag : int
        Flag to check in each integration.

    set_flag : int
        Flag to set if flag is found in each integration.
    """
    # Checking for all groups is the same as checking for all integrations
    # because in both we are checking cubes.  For the integration check the
    # first dimension is the number of integrations, for the ramp check the
    # first dimension is the number of groups.
    set_if_total_integ(pixeldq_sect, gdq_sect, flag, set_flag)


def compute_median_rates(ramp_data):
    """
    Compute the median rates for each pixel.

    Parameter
    ---------
    ramp_data : RampData
        Contains the data needed to compute the median rates.

    Return
    ------
    median_rates : ndarray
        The computed median rates for each pixel.
    """
    nints, ngroups, nrows, ncols = ramp_data.data.shape
    imshape = (nrows, ncols)

    group_time = ramp_data.group_time
    frame_time = ramp_data.frame_time
    adjustment = group_time / frame_time

    if ramp_data.one_groups_time is not None:
        one_groups_time_adjustment = group_time / ramp_data.one_groups_time
    else:
        one_groups_time_adjustment = None

    median_diffs_2d = np.zeros(imshape, dtype=np.float32)

    for integ in range(nints):
        data_sect = ramp_data.data[integ, :, :, :].copy()
        gdq_sect = ramp_data.groupdq[integ, :, :, :]

        # Reset all saturated groups in the input data array to NaN
        # data_sect[np.bitwise_and(gdq_sect, ramp_data.flags_saturated).astype(bool)] = np.nan
        invalid_flags = ramp_data.flags_saturated | ramp_data.flags_do_not_use
        invalid_locs = np.bitwise_and(gdq_sect, invalid_flags).astype(bool)
        data_sect[invalid_locs] = np.nan
        data_sect = data_sect / group_time

        if one_groups_time_adjustment is not None:
            one_groups_locs = ramp_data.one_groups_locs[integ]
            tmp_dsect = data_sect[0, :, :]
            tmp_dsect[one_groups_locs] = tmp_dsect[one_groups_locs] * one_groups_time_adjustment
            data_sect[0, :, :] = tmp_dsect

        if ramp_data.zframe_locs is not None:
            zinteg_locs = ramp_data.zframe_locs[integ]
            tmp_dsect = data_sect[0, :, :]
            tmp_dsect[zinteg_locs] = tmp_dsect[zinteg_locs] * adjustment
            data_sect[0, :, :] = tmp_dsect

        # Compute the first differences of all groups
        first_diffs_sect = np.diff(data_sect, axis=0)

        # If the dataset has only 1 group/integ, assume the 'previous group'
        #   is all zeros, so just use data as the difference
        if first_diffs_sect.shape[0] == 0:
            first_diffs_sect = data_sect.copy()
        else:
            # Similarly, for datasets having >1 group/integ and having
            #   single-group segments, just use the data as the difference
            wh_nan = np.where(np.isnan(first_diffs_sect[0, :, :]))

            if len(wh_nan[0]) > 0:
                first_diffs_sect[0, :, :][wh_nan] = data_sect[0, :, :][wh_nan]

            del wh_nan

            # Mask all the first differences that are affected by a CR,
            #   starting at group 1.  The purpose of starting at index 1 is
            #   to shift all the indices down by 1, so they line up with the
            #   indices in first_diffs.
            (
                i_group,
                i_yy,
                i_xx,
            ) = np.where(np.bitwise_and(gdq_sect[1:, :, :], ramp_data.flags_jump_det))
            first_diffs_sect[i_group, i_yy, i_xx] = np.nan

            del i_group, i_yy, i_xx

            # Check for pixels in which there is good data in 0th group, but
            #   all first_diffs for this ramp are NaN because there are too
            #   few good groups past the 0th. Due to the shortage of good
            #   data, the first_diffs will be set here equal to the data in
            #   the 0th group.
            wh_min = np.where(
                np.logical_and(np.isnan(first_diffs_sect).all(axis=0), np.isfinite(data_sect[0, :, :]))
            )
            if len(wh_min[0] > 0):
                first_diffs_sect[0, :, :][wh_min] = data_sect[0, :, :][wh_min]

            del wh_min


        # All first differences affected by saturation and CRs have been set
        #  to NaN, so compute the median of all non-NaN first differences.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "All-NaN.*", RuntimeWarning)
            nan_med = np.nanmedian(first_diffs_sect, axis=0)

        nan_med[np.isnan(nan_med)] = 0.0  # if all first_diffs_sect are nans
        median_diffs_2d[:, :] += nan_med

    # Compute the final 2D array of differences; create rate array
    median_rates = median_diffs_2d / nints
    del median_diffs_2d
    del data_sect

    return median_rates


def use_zeroframe_for_saturated_ramps(ramp_data):
    """
    For saturated ramps, if there is good data in the ZEROFRAME, replace
    group zero with the data in ZEROFRAME to use the ramp as a one group
    ramp.

    Parameters
    ----------
    model : data model
        input data model, assumed to be of type RampModel

    Return
    ------
    zframe_locs : list
        A 2D list for the location of the ramps using ZEROFRAME data.
        zframe_locs[k] is the list of pixels in the kth integration.
    """
    nints, ngroups, nrows, ncols = ramp_data.data.shape
    sat_flag = ramp_data.flags_saturated
    good_flag = 0
    dq = ramp_data.groupdq

    zframe_locs = [None] * nints

    cnt = 0
    zframe_mat = np.zeros((nints, nrows, ncols), dtype=np.uint8)
    for integ in range(nints):
        intdq = dq[integ, :, :, :]

        # Find ramps with a good zeroeth group, but saturated in
        # the remainder of the ramp.
        wh_sat = groups_saturated_in_integration(intdq, sat_flag, ngroups)

        whs_rows, whs_cols = wh_sat
        row_list = []
        col_list = []
        for n in range(len(whs_rows)):
            row = whs_rows[n]
            col = whs_cols[n]

            # For ramps completely saturated look for data in the ZEROFRAME
            # that is non-zero.  If it is non-zero, replace group zero in the
            # ramp with the data in ZEROFRAME.
            if ramp_data.zeroframe[integ, row, col] != 0:
                row_list.append(row)
                col_list.append(col)
                ramp_data.data[integ, 0, row, col] = ramp_data.zeroframe[integ, row, col]
                ramp_data.groupdq[integ, 0, row, col] = good_flag
                zframe_mat[integ, row, col] = 1
                cnt = cnt + 1

        zframe_locs[integ] = (np.array(row_list, dtype=int), np.array(col_list, dtype=int))

    return zframe_mat, zframe_locs, cnt


def groups_saturated_in_integration(intdq, sat_flag, num_sat_groups):
    """
    Find the ramps in an integration that have num_sat_groups saturated.

    Parameters
    ----------
    intdq : ndarray
        DQ flags for an integration

    sat_flag : uint
        The data quality flag for SATURATED

    num_sat_groups : int
        The number of saturated groups in an integration of interest.
    """
    sat_groups = np.zeros(intdq.shape, dtype=int)
    sat_groups[np.bitwise_and(intdq, sat_flag).astype(bool)] = 1
    nsat_groups = sat_groups.sum(axis=0)
    return np.where(nsat_groups == num_sat_groups)



import importlib
import sys
from datetime import datetime
from pathlib import Path

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib


REPO_ROOT = Path(__file__).parent.parent

# Modules that automodapi will document need to be available
# in the path:
sys.path.insert(0, str(REPO_ROOT / "src" / "stcal"))

# Read the package's `pyproject.toml` so that we can use relevant
# values here:
with (REPO_ROOT / "pyproject.toml").open("rb") as configuration_file:
    conf = tomllib.load(configuration_file)
setup_metadata = conf["project"]

project = setup_metadata["name"]
primary_author = setup_metadata["authors"][0]
author = f'{primary_author["name"]} <{primary_author["email"]}>'
copyright = f'{datetime.now().year}, {primary_author["name"]}'  # noqa: A001

package = importlib.import_module(project)
version = package.__version__.split("-", 1)[0]
release = package.__version__

# Configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/devdocs", None),
    "scipy": ("http://scipy.github.io/devdocs", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "gwcs": ("https://gwcs.readthedocs.io/en/latest/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "tweakwcs": ("https://tweakwcs.readthedocs.io/en/latest/", None),
}

nitpick_ignore = [("py:class", "optional"), ("py:class", "np.ndarray")]

extensions = [
    "pytest_doctestplus.sphinx.doctestplus",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.automodsumm",
    "sphinx_automodapi.autodoc_enhancements",
    "sphinx_automodapi.smart_resolver",
    "sphinx_asdf",
    "sphinx.ext.mathjax",
]

autosummary_generate = True
numpydoc_show_class_members = False
autoclass_content = "both"

html_theme = "sphinx_rtd_theme"
html_logo = "_static/stsci_pri_combo_mark_white.png"
html_theme_options = {
    "collapse_navigation": True,
}
html_domain_indices = True
html_sidebars = {"**": ["globaltoc.html", "relations.html", "searchbox.html"]}
html_use_index = True

# Enable nitpicky mode - which ensures that all references in the docs resolve.
nitpicky = True



INDENT = "    "

class RampData:
    def __init__(self):
        """Creates an internal ramp fit class."""
        # Arrays from the data model
        self.data = None
        self.err = None
        self.groupdq = None
        self.pixeldq = None
        self.average_dark_current = None

        # Needed for CHARGELOSS recomputation
        self.orig_gdq = None
        self.algorithm = None

        # Meta information
        self.instrument_name = None
        self.read_pattern = None
        self.rejection_threshold = None

        self.frame_time = None
        self.group_time = None
        self.groupgap = None
        self.nframes = None
        self.drop_frames1 = None

        # Data quality flags
        self.flags_do_not_use = None
        self.flags_jump_det = None
        self.flags_saturated = None
        self.flags_no_gain_val = None
        self.flags_unreliable_slope = None
        self.flags_chargeloss = None

        # ZEROFRAME
        self.zframe_mat = None
        self.zframe_locs = None
        self.zframe_cnt = 0
        self.zeroframe = None

        # Slice info
        self.start_row = None
        self.num_rows = None

        # One group ramp suppression for saturated ramps after 0th group.
        self.suppress_one_group_ramps = False

        # C code debugging switch.
        self.run_c_code = False

        self.one_groups_locs = None  # One good group locations.
        self.one_groups_time = None  # Time to use for one good group ramps.

        self.current_integ = -1

        self.debug = False

    def set_arrays(self, data, err, groupdq, pixeldq, average_dark_current, orig_gdq=None):
        """
        Set the arrays needed for ramp fitting.

        Parameter
        ---------
        data : ndarray
            4-D array containing the pixel information.  It has dimensions
            (nintegrations, ngroups, nrows, ncols)

        err : ndarray
            4-D array containing the error information.  It has dimensions
            (nintegrations, ngroups, nrows, ncols)

        groupdq : ndarray (uint16)
            4-D array containing the data quality flags.  It has dimensions
            (nintegrations, ngroups, nrows, ncols)

        pixeldq : ndarray (uint32)
            4-D array containing the pixel data quality information.  It has dimensions
            (nintegrations, ngroups, nrows, ncols)

        average_dark_current : ndarray (float32)
            2-D array containing the average dark current. It has
            dimensions (nrows, ncols)

        orig_gdq : ndarray
            4-D array containing a copy of the original group DQ array.  Since
            the group DQ array can be modified during ramp fitting, this keeps
            around the original group DQ flags passed to ramp fitting.
        """
        # Get arrays from the data model
        self.data = data
        self.err = err
        self.groupdq = groupdq
        self.pixeldq = pixeldq
        self.average_dark_current = average_dark_current

        self.orig_gdq = orig_gdq

    def set_meta(self, name, frame_time, group_time, groupgap, nframes, drop_frames1=None):
        """
        Set the metainformation needed for ramp fitting.

        Parameter
        ---------
        name : str
            The instrument name.

        frame_time : float32
            The time to read one frame.

        group_time : float32
            The time to read one group.

        groupgap : int
            The number of frames that are not included in the group average

        nframes : int
            The number of frames that are included in the group average

        drop_frames1 :
            The number of frames dropped at the beginning of every integration.
            May not be used in some pipelines, so is defaulted to NoneType.
        """
        # Get meta information
        self.instrument_name = name

        self.frame_time = frame_time
        self.group_time = group_time
        self.groupgap = groupgap
        self.nframes = nframes

        # May not be available for all pipelines, so is defaulted to NoneType.
        self.drop_frames1 = drop_frames1

    def set_dqflags(self, dqflags):
        """
        Set the data quality flags needed for ramp fitting.

        Parameter
        ---------
        dqflags : dict
            A dictionary with specific key words needed for processing.
        """
        # Get data quality flags
        self.flags_do_not_use = dqflags["DO_NOT_USE"]
        self.flags_jump_det = dqflags["JUMP_DET"]
        self.flags_saturated = dqflags["SATURATED"]
        self.flags_no_gain_val = dqflags["NO_GAIN_VALUE"]
        self.flags_unreliable_slope = dqflags["UNRELIABLE_SLOPE"]
        if self.algorithm is not None and self.algorithm.upper() == "OLS_C":
            self.flags_chargeloss = dqflags["CHARGELOSS"]

    def dbg_print_types(self):
        # Arrays from the data model
        print("-" * 80)
        print("    Array Types:")
        print(f"data : {type(self.data)}")
        print(f"err : {type(self.err)}")
        print(f"groupdq : {type(self.groupdq)}")
        print(f"pixeldq : {type(self.pixeldq)}")

        self.dbg_print_meta()
        self.dbg_print_mp()
        self.dbg_print_zframe()
        self.dbg_print_1grp()

    def dbg_print_meta(self):
        # Meta information
        print("-" * 80)
        print("    Meta:")
        print(f"Instrument: {self.instrument_name}")

        print(f"Frame time : {self.frame_time}")
        print(f"Group time : {self.group_time}")
        print(f"Group Gap : {self.groupgap}")
        print(f"Nframes : {self.nframes}")
        print(f"Drop Frames : {self.drop_frames1}")
        print("-" * 80)

    def dbg_print_mp(self):
        # Multiprocessing
        print("-" * 80)
        print(f"Start row : {self.start_row}")
        print(f"Number of rows : {self.num_rows}")

    def dbg_print_zframe(self):
        # ZEROFRAME
        print("-" * 80)
        print("    ZEROFRAME:")
        print(f"zframe_locs : {type(self.zframe_locs)}")
        print(f"zeroframe : {type(self.zeroframe)}")

    def dbg_print_1grp(self):
        # One group ramp suppression for saturated ramps after 0th group.
        print("-" * 80)
        print("    One Group Suppression:")
        print(f"suppress_one_group_ramps : {type(self.suppress_one_group_ramps)}")

    def dbg_print_basic_info(self):
        # Arrays from the data model
        self.dbg_print_meta()

        print(f"Shape : {self.data.shape}")
        print(f"data : \n{self.data}")
        print(f"groupdq : \n{self.groupdq}")
        # print(f"err : \n{self.err}")
        # print(f"pixeldq : \n{self.pixeldq}")
        print("-" * 80)

    def dbg_print_pixel_info(self, row, col):
        print("-" * 80)
        print(f"    data")
        for integ in range(self.data.shape[0]):
            print(f"[{integ}] {self.data[integ, :, row, col]}")
        print(f"    groupdq")
        for integ in range(self.data.shape[0]):
            print(f"[{integ}] {self.groupdq[integ, :, row, col]}")
        # print(f"    err :\n{self.err[:, :, row, col]}")
        # print(f"    pixeldq :\n{self.pixeldq[row, col]}")

    def dbg_print_info(self):
        print(" ")
        nints, ngroups, nrows, ncols = self.data.shape
        for row in range(nrows):
            for col in range(ncols):
                print("=" * 80)
                print(f"**** Pixel ({row}, {col}) ****")
                self.dbg_print_pixel_info(row, col)
        print("=" * 80)

    def dbg_write_ramp_data_pix_pre(self, fname, row, col, fd):
        fd.write("def create_ramp_data_pixel():\n")
        indent = INDENT
        fd.write(f"{indent}'''\n")
        fd.write(f"{indent}Using pixel ({row}, {col})\n")
        fd.write(f"{indent}'''\n")
        fd.write(f"{indent}ramp_data = RampData()\n\n")

        fd.write(f"{indent}ramp_data.instrument_name = '{self.instrument_name}'\n\n")

        fd.write(f"{indent}ramp_data.frame_time = {self.frame_time}\n")
        fd.write(f"{indent}ramp_data.group_time = {self.group_time}\n")
        fd.write(f"{indent}ramp_data.groupgap = {self.groupgap}\n")
        fd.write(f"{indent}ramp_data.nframes = {self.nframes}\n")
        fd.write(f"{indent}ramp_data.drop_frames1 = {self.drop_frames1}\n\n")

        fd.write(f"{indent}ramp_data.flags_do_not_use = {self.flags_do_not_use}\n")
        fd.write(f"{indent}ramp_data.flags_jump_det = {self.flags_jump_det}\n")
        fd.write(f"{indent}ramp_data.flags_saturated = {self.flags_saturated}\n")
        fd.write(f"{indent}ramp_data.flags_no_gain_val = {self.flags_no_gain_val}\n")
        fd.write(f"{indent}ramp_data.flags_unreliable_slope = {self.flags_unreliable_slope}\n\n")


        fd.write(f"{indent}ramp_data.start_row = 0\n")
        fd.write(f"{indent}ramp_data.num_rows = 1\n\n")

        fd.write(f"{indent}ramp_data.suppress_one_group_ramps = {self.suppress_one_group_ramps}\n\n")

        nints, ngroups, nrows, ncols = self.data.shape
        fd.write(f"{indent}data = np.zeros(({nints}, {ngroups}, 1, 1), dtype=np.float32)\n")
        fd.write(f"{indent}err = np.zeros(({nints}, {ngroups}, 1, 1), dtype=np.float32)\n")
        fd.write(f"{indent}gdq = np.zeros(({nints}, {ngroups}, 1, 1), dtype=np.uint8)\n")
        fd.write(f"{indent}pdq = np.zeros((1, 1), dtype=np.uint32)\n")


    def dbg_write_ramp_data_pix_post(self, fname, row, col, fd):
        indent = INDENT

        fd.write(f"{indent}ramp_data.data = data\n")
        fd.write(f"{indent}ramp_data.err = err\n")
        fd.write(f"{indent}ramp_data.groupdq = gdq\n")
        fd.write(f"{indent}ramp_data.pixeldq = pdq\n")
        fd.write(f"{indent}ramp_data.zeroframe = zframe\n\n")

        fd.write(f"{indent}return ramp_data, ngain, nrnoise\n")

    def dbg_write_ramp_data_pix_pixel(self, fname, row, col, gain, rnoise, fd):
        import numpy as np
        indent = INDENT

        # XXX Make this a separate function
        delimiter = "-" * 40
        fd.write(f"{indent}# {delimiter}\n\n");
        fd.write(f"{indent}# ({row}, {col})\n\n");

        nints = self.data.shape[0]

        for integ in range(nints):
            arr_str = np.array2string(self.data[integ, :, row, col], precision=12, max_line_width=np.nan, separator=", ")
            fd.write(f"{indent}data[{integ}, :, 0, 0] = np.array({arr_str})\n")
        fd.write("\n")

        for integ in range(nints):
            arr_str = np.array2string(self.err[integ, :, row, col], precision=12, max_line_width=np.nan, separator=", ")
            fd.write(f"{indent}err[{integ}, :, 0, 0] = np.array({arr_str})\n")
        fd.write("\n")

        for integ in range(nints):
            arr_str = np.array2string(self.groupdq[integ, :, row, col], precision=12, max_line_width=np.nan, separator=", ")
            fd.write(f"{indent}gdq[{integ}, :, 0, 0] = np.array({arr_str})\n")
        fd.write("\n")

        arr_str = np.array2string(self.pixeldq[row, col], precision=12, max_line_width=np.nan, separator=", ")
        fd.write(f"{indent}pdq[0, 0] = {arr_str}\n\n")

        if self.zeroframe is not None:
            fd.write(f"{indent}zframe = np.zeros((1, 1), dtype=np.float32)\n\n")
            arr_str = np.array2string(self.zeroframe[row, col], precision=12, max_line_width=np.nan, separator=", ")
            fd.write(f"{indent}zframe[0, 0] = {arr_str}\n\n")
        else:
            fd.write(f"{indent}zframe = None\n\n")

        fd.write(f"{indent}ngain = np.zeros((1, 1), dtype=np.float32)\n")
        fd.write(f"{indent}ngain[0, 0] = {gain[row, col]}\n\n")

        fd.write(f"{indent}nrnoise = np.zeros((1, 1), dtype=np.float32)\n")
        fd.write(f"{indent}nrnoise[0, 0] = {rnoise[row, col]}\n\n")


    def dbg_write_ramp_data_pix(self, fname, row, col, gain, rnoise):
        print(f"*** {fname} ***")
        with open(fname, "w") as fd:
            self.dbg_write_ramp_data_pix_pre(fname, row, col, fd)
            self.dbg_write_ramp_data_pix_pixel(fname, row, col, gain, rnoise, fd)
            self.dbg_write_ramp_data_pix_post(fname, row, col, fd)



import warnings

warnings.warn(
    "basic_utils has been moved to stdatamodels.basic_utils, please use that instead",
    DeprecationWarning,
    stacklevel=2,
)

try:
    from stdatamodels.basic_utils import multiple_replace
except ImportError as err:
    msg = "basic_utils has been moved to stdatamodels.basic_utils, please install stdatamodels"
    raise ImportError(msg) from err


__all__ = ["multiple_replace"]
